"""Polynomial machine learning potential interface."""

# Copyright (C) 2024 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import os
from collections.abc import Sequence
from dataclasses import dataclass
from typing import Any, Optional, Union

import numpy as np

from phonopy.file_IO import get_io_module_to_decompress
from phonopy.structure.atoms import PhonopyAtoms

try:
    from pypolymlp.mlp_dev.pypolymlp import Pypolymlp
except ImportError:
    Pypolymlp = Any


@dataclass
class PypolymlpParams:
    """Parameters for pypolymlp.

    cutoff : flaot, optional
        Cutoff radius. Default is 8.0.
    model_type : int, optional
        Polynomial function type. Default is 3. model_type = 1: Linear
        polynomial of polynomial invariants model_type = 2: Polynomial of
        polynomial invariants model_type = 3: Polynomial of pair invariants
                        + linear polynomial of polynomial invariants
        model_type = 4: Polynomial of pair and second-order invariants
                        + linear polynomial of polynomial invariants
    max_p : int, optional
        Order of polynomial function. Default is 2.
    gtinv_order : int, optional
        Maximum order of polynomial invariants. Default is 3.
    gtinv_maxl: Sequence[int], optional
        Maximum angular numbers of polynomial invariants. [maxl for order=2,
        maxl for order=3, ...] Default is (8, 8).
    gaussian_params1, gaussian_params2 : Sequence[float, float, int], optional
        Parameters for exp[- param1 * (r - param2)**2]. Parameters are given as
        np.linspace(p[0], p[1], p[2]), where p[0], p[1], and p[2] are given by
        gaussian_params1 and gaussian_params2. Normally it is recommended to
        modify only gaussian_params2. Default is (1.0, 1.0, 1) and (0.0, 7.0,
        10), respectively.
    atom_energies: dict[str, float], optional
        Atomic energies specified by dictionary, e.g., {'Si': -0.35864636, 'O':
        -0.95743902}, where the order is irrelevant. Default is None, which
        gives zero energies for all atoms.

    """

    cutoff: float = 8.0
    model_type: int = 3
    max_p: int = 2
    gtinv_order: int = 3
    gtinv_maxl: Sequence[int] = (8, 8)
    gaussian_params1: Sequence[float, float, int] = (1.0, 1.0, 1)
    gaussian_params2: Sequence[float, float, int] = (0.0, 7.0, 10)
    atom_energies: Optional[dict[str, float]] = None
    ntrain: Optional[int] = None
    ntest: Optional[int] = None


@dataclass
class PypolymlpData:
    """Dataset for pypolymlp input.

    displacements : np.ndarray
        Displacements of atoms. shape=(n, natoms, 3)
    forces : np.ndarray
        Displacements of atoms. shape=(n, natoms, 3)
    supercell_energies : np.ndarray, optional
        Energies of supercells. shape=(n,)

    """

    displacements: np.ndarray
    forces: np.ndarray
    supercell_energies: np.ndarray


def develop_pypolymlp(
    supercell: PhonopyAtoms,
    train_data: PypolymlpData,
    test_data: PypolymlpData,
    params: Optional[PypolymlpParams] = None,
    verbose: bool = False,
) -> Pypolymlp:  # type: ignore
    """Develop polynomial MLPs of pypolymlp.

    Parameters
    ----------
    supercell : PhonopyAtoms
        Supercell structure.
    train_data : PyPolymlpData
        Training dataset.
    test_data : PyPolymlpData
        Test dataset.
    params : PypolymlpParams, optional
        Parameters for pypolymlp. Default is None.
    verbose : bool, optional
        Verbosity. Default is False.

    Returns
    -------
    polymlp : Pypolymlp
        Pypolymlp object.

    """
    try:
        from pypolymlp.mlp_dev.pypolymlp import Pypolymlp
        from pypolymlp.utils.phonopy_utils import phonopy_cell_to_structure
    except ImportError as exc:
        raise ModuleNotFoundError("Pypolymlp python module was not found.") from exc

    if params is None:
        _params = PypolymlpParams()
    else:
        _params = params

    if _params.atom_energies is None:
        elements_energies = {s: 0.0 for s in supercell.symbols}
    else:
        elements_energies = {s: _params.atom_energies[s] for s in supercell.symbols}
    polymlp = Pypolymlp()
    polymlp.set_params(
        elements=list(elements_energies.keys()),
        cutoff=_params.cutoff,
        model_type=_params.model_type,
        max_p=_params.max_p,
        gtinv_order=_params.gtinv_order,
        gtinv_maxl=_params.gtinv_maxl,
        gaussian_params2=_params.gaussian_params2,
        atomic_energy=list(elements_energies.values()),
    )
    polymlp.set_datasets_displacements(
        train_data.displacements.transpose(0, 2, 1),
        train_data.forces.transpose(0, 2, 1),
        train_data.supercell_energies,
        test_data.displacements.transpose(0, 2, 1),
        test_data.forces.transpose(0, 2, 1),
        test_data.supercell_energies,
        phonopy_cell_to_structure(supercell),
    )
    polymlp.run(verbose=verbose)
    return polymlp


def evalulate_pypolymlp(
    polymlp: Pypolymlp,  # type: ignore
    supercells_with_displacements: list[PhonopyAtoms],
) -> list[np.ndarray, np.ndarray, np.ndarray]:
    """Run force calculation using pypolymlp.

    Parameters
    ----------
    polymlp : Pypolymlp
        Pypolymlp object.
    supercells_with_displacements : Sequence[PhonopyAtoms]
        Sequence of supercells with displacements.

    Returns
    -------
    energies : np.ndarray
        Energies of supercells. shape=(n,)
    forces : np.ndarray
        Forces of supercells. shape=(n, natoms, 3)
    stresses : np.ndarray
        Stresses of supercells (xx, yy, zz, xy, yz, zx). shape=(n, 6)

    """
    try:
        from pypolymlp.calculator.properties import Properties
        from pypolymlp.utils.phonopy_utils import phonopy_cell_to_structure
    except ImportError as exc:
        raise ModuleNotFoundError("Pypolymlp python module was not found.") from exc

    prop = Properties(params=polymlp.parameters, coeffs=polymlp.coeffs)
    energies, forces, stresses = prop.eval_multiple(
        [phonopy_cell_to_structure(scell) for scell in supercells_with_displacements]
    )
    energies = np.array(energies, dtype="double")
    forces = np.array(np.transpose(forces, (0, 2, 1)), dtype="double", order="C")
    stresses = np.array(stresses, dtype="double", order="C")
    return energies, forces, stresses


def parse_mlp_params(params: Union[str, dict, PypolymlpParams]) -> PypolymlpParams:
    """Parse MLP parameters string and return PypolymlpParams.

    Supported MLP parameters
    ------------------------
    cutoff: float = 8.0
    model_type: int = 3
    max_p: int = 2
    gtinv_order: int = 3
    gtinv_maxl: Sequence[int] = (8, 8)
    gaussian_params1: Sequence[float, float, int] = (1.0, 1.0, 1)
    gaussian_params2: Sequence[float, float, int] = (0.0, 7.0, 10)
    atom_energies: Optional[dict[str, float]] = None
    ntrain: Optional[int] = None
    ntest: Optional[int] = None

    Parameters
    ----------
    params : str, dict, PyPolymlpParams
        Parameters for pypolymlp.

    Note
    ----
    When str, it should be written as follows:

        "cutoff = 10.0, gtinv_maxl = 8 8"
        "atom_energies = Si -0.35864636 O -0.95743902"


    """
    if isinstance(params, dict):
        return PypolymlpParams(**params)
    elif isinstance(params, PypolymlpParams):
        return params
    elif isinstance(params, str):
        params_dict = {}
        for param in params.split(","):
            key_val = [v.strip().lower() for v in param.split("=")]
            if len(key_val) != 2:
                break
            key, val = key_val
            if key == "gtinv_maxl":
                params_dict[key] = tuple(map(int, val.split()))
            elif key == "gaussian_params1" or key == "gaussian_params2":
                vals = val.split()
                params_dict[key] = (float(vals[0]), float(vals[1]), int(vals[2]))
            elif key == "atom_energies":
                vals = val.split()
                if len(vals) % 2 != 0:
                    raise ValueError(
                        "The input list must have an even number of elements."
                    )
                params_dict[key] = {
                    vals[i]: float(vals[i + 1]) for i in range(0, len(vals), 2)
                }
            elif key == "cutoff":
                params_dict[key] = float(val)
            else:
                if key in ("model_type", "max_p", "gtinv_order", "ntrain", "ntest"):
                    params_dict[key] = int(val)
        return PypolymlpParams(**params_dict)
    else:
        raise RuntimeError("params has to be dict, str, or PypolymlpParams.")


def save_pypolymlp(mlp: Pypolymlp, filename: str):  # type: ignore
    """Save MLP data to file."""
    mlp.save_mlp(filename=filename)


def load_pypolymlp(filename: Optional[Union[str, bytes, os.PathLike]]) -> Pypolymlp:  # type: ignore
    """Load MLP data from file."""
    mlp = Pypolymlp()
    myio = get_io_module_to_decompress(filename)
    with myio.open(filename, "rt") as fp:
        mlp.load_mlp(fp)
    return mlp


def develop_mlp_by_pypolymlp(
    mlp_dataset: dict,
    supercell: PhonopyAtoms,
    params: Optional[Union[PypolymlpParams, dict, str]] = None,
    test_size: float = 0.1,
    log_level: int = 0,
) -> Pypolymlp:  # type: ignore
    """Develop MLPs by pypolymlp."""
    if params is not None:
        _params = parse_mlp_params(params)
    else:
        _params = params

    if _params is not None and _params.ntrain is not None and _params.ntest is not None:
        ntrain = _params.ntrain
        ntest = _params.ntest
        disps = mlp_dataset["displacements"]
        forces = mlp_dataset["forces"]
        energies = mlp_dataset["supercell_energies"]
        train_data = PypolymlpData(
            displacements=disps[:ntrain],
            forces=forces[:ntrain],
            supercell_energies=energies[:ntrain],
        )
        test_data = PypolymlpData(
            displacements=disps[-ntest:],
            forces=forces[-ntest:],
            supercell_energies=energies[-ntest:],
        )
    else:
        disps = mlp_dataset["displacements"]
        forces = mlp_dataset["forces"]
        energies = mlp_dataset["supercell_energies"]
        n = int(len(disps) * (1 - test_size))
        train_data = PypolymlpData(
            displacements=disps[:n],
            forces=forces[:n],
            supercell_energies=energies[:n],
        )
        test_data = PypolymlpData(
            displacements=disps[n:],
            forces=forces[n:],
            supercell_energies=energies[n:],
        )
    mlp = develop_pypolymlp(
        supercell,
        train_data,
        test_data,
        params=_params,
        verbose=log_level - 1 > 0,
    )
    return mlp



"""Calculate random displacements from phonons at finite temperatures."""

# Copyright (C) 2018 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

from collections.abc import Sequence
from typing import Optional, Union

import numpy as np

from phonopy.harmonic.dynamical_matrix import get_dynamical_matrix
from phonopy.harmonic.dynmat_to_fc import (
    DynmatToForceConstants,
    categorize_commensurate_points,
    get_commensurate_points_in_integers,
)
from phonopy.physical_units import get_physical_units
from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.cells import Primitive


def bose_einstein_dist(
    x: Union[np.ndarray, float], t: float
) -> Union[np.ndarray, float]:
    """Return Bose-Einsetein distribution.

    Parameters
    ----------
    x : ndarray or float
        Phonon frequencies in THz.
    t : float
        Temperature in K.

    Returns
    -------
    ndarray or float
        Phonon occupation numbers.

    """
    return 1.0 / (
        np.exp(get_physical_units().THzToEv * x / (get_physical_units().KB * t)) - 1
    )


class RandomDisplacements:
    """Generate random displacements by Canonical ensenmble.

    Note
    ----
    Phonon frequencies are used to calculate phonon occupation number,
    for which phonon frequencies have to be given in THz. Therefore unit
    conversion factor has to be specified at the initialization.

    Imaginary phonon modes are treated so as to have their absolute phonon
    frequencies |omega| and phonon modes having |omega| < cutoff_frequency
    are ignored.

    Attributes
    ----------
    u : ndarray
        Random atomic displacements generated by canonical distribution of
        harmonic oscillator. The unit of distance is Angstrom.
        shape=(number_of_snapshots, supercell_atoms, 3)
        dtype='double', order='C'
    qpoints : ndarray
        Commensurate q-points corresponding to the supercell matrix but not
        all. Only half of the commensurate q-points that are not on the BZ
        boundary and Gamma-points are only taken, because of the symmetry
        of dynamical matrix: omega_q = omega_-q and e_q = e_-q^*.
    frequencies : ndarray
        Phonon frequencies at commensurate q-points as explained above
        qpoints attribute. Both of getter and setter are implemented.
        The aim of this is to modify random displacements by modifying
        frequencies by users. The unit is in THz.
        shape=(len(qpoints), num_band), dtype='double', order='C'
        where num_band is 3 * number of atoms in primitive cell.
    force_constants : ndarray
        Force constants calculated from phonon frequencies and eigenvectors
        at commensurate q-points as given above qpoints attribute. By this,
        phonon can be calculated with modified phonon frequencies. To
        calculate force constants, run_d2f has to be executed. For example,

            rd = RandomDisplacements(supercell, primitive, force_constants)
            freqs = rd.frequencies
            ... modify freqs by users
            rd.frequencies = freqs
            rd.run(500)  # To get random displacements
            rd.run_d2f()
            fc = rd.force_constants  # To draw phonons with modified freqs

        shape=(superell_atoms, supercell_atoms, 3, 3)
        dtype='double', order='C'

    """

    def __init__(
        self,
        supercell: PhonopyAtoms,
        primitive: Primitive,
        force_constants: Union[np.ndarray, Sequence],
        dist_func: Optional[str] = None,
        cutoff_frequency: Optional[float] = None,
        max_distance: Optional[float] = None,
        factor: Optional[float] = None,
        use_openmp: bool = False,
    ):
        """Init method.

        Parameters
        ----------
        supercell : PhonopyAtoms
            Supercell.
        primitive : Primitive
            Primitive cell
        force_constants : array_like
            Force constants matrix. See the details at docstring of
            DynamialMatrix.
        dist_func : str or None
            Harmonic oscillator distribution function either by 'quantum'
            or 'classical'. The starndard deviation of normal distribution
            is determined following the choice. Default is None, corresponding
            to 'quantum'.
        cutoff_frequency : float
            Lowest phonon frequency below which frequency the phonon mode
            is treated specially. See _get_sigma. Default is None, which
            means 0.01.
        max_distance : float or None, optional
            In random displacements generation from canonical ensemble of
            harmonic phonons, displacements larger than max distance are
            renormalized to the max distance, i.e., a disptalcement d is shorten
            by d -> d / |d| * max_distance if |d| > max_distance.
        factor : float
            Phonon frequency unit conversion factor to THz
        use_openmp : bool, optional, default=False
            Use OpenMP in calculate dynamical matrix and its inverse.

        """
        if cutoff_frequency is None or cutoff_frequency < 0:
            self._cutoff_frequency = 0.01
        else:
            self._cutoff_frequency = cutoff_frequency
        self._max_distance = max_distance

        _physical_units = get_physical_units()
        if factor is None:
            self._factor = _physical_units.DefaultToTHz
        else:
            self._factor = factor
        self._T = None
        self._u = None

        if dist_func is None or dist_func == "quantum":
            self._dist_func = "quantum"
        elif dist_func == "classical":
            self._dist_func = "classical"
        else:
            raise RuntimeError("Either 'quantum' or 'classical' is required.")

        self._unit_conversion = (
            _physical_units.Hbar
            * _physical_units.EV
            / _physical_units.AMU
            / _physical_units.THz
            / (2 * np.pi)
            / _physical_units.Angstrom**2
        )
        self._unit_conversion_classical = (
            _physical_units.KB
            * _physical_units.EV
            / _physical_units.AMU
            / (_physical_units.THz * (2 * np.pi)) ** 2
            / _physical_units.Angstrom**2
        )

        # Dynamical matrix without NAC because of commensurate points only
        self._dynmat = get_dynamical_matrix(
            force_constants, supercell, primitive, use_openmp=use_openmp
        )

        self._comm_points = None
        self._ii = None
        self._ij = None
        self._setup_sampling_qpoints(supercell.cell, primitive.cell)

        s2p = primitive.s2p_map
        p2p = primitive.p2p_map
        self._s2pp = [p2p[i] for i in s2p]
        # Transformation matrix of scaled supercell positions to primitive
        tmat = np.dot(supercell.cell, np.linalg.inv(primitive.cell))
        self._spos = np.dot(self._dynmat.supercell.scaled_positions, tmat)
        self._ppos = self._dynmat.primitive.scaled_positions
        self._lpos = self._spos - self._ppos[self._s2pp]

        self._eigvals_ii = None
        self._eigvecs_ii = None
        self._phase_ii = None
        self._eigvals_ij = None
        self._eigvecs_ij = None
        self._phase_ij = None
        self._prepare()

        # This is set when running run_d2f.
        # The aim is to produce force constants from modified frequencies.
        self._force_constants = None

        # Displacement correlation matrix (nsatom, nsatom, 3, 3)
        self._uu = None
        self._uu_inv = None

        # Phonon bands included in integration of dispalcements.
        self._conditions_ii = None
        self._conditions_ij = None

    def run(
        self,
        T: float,
        number_of_snapshots: int = 1,
        random_seed: Optional[int] = None,
        randn: Optional[tuple] = None,
    ):
        """Calculate random displacements.

        Random displacements are calculated from eigensolutions stored by
        self._prepare() method.

        Parameters
        ----------
        T : float
            Temperature in Kelvin.
        number_of_snapshots : int, optional
            Number of snapshots to be generated. Default is 1.
        random_seed : int or None, optional
            Random seed passed to np.random.default_rng(seed). Default is None.
        randn : tuple
            (randn_ii, randn_ij). Used for testing purpose for the fixed random
            numbers of random.Generator.standard_normal that can depends on
            system.

        """
        if np.issubdtype(type(random_seed), np.integer):
            rng = np.random.default_rng(seed=random_seed)
        else:
            rng = np.random.default_rng()

        N = len(self._comm_points)

        # This randn is used only for testing purpose.
        if randn is None:
            shape = (
                len(self._eigvals_ii),
                number_of_snapshots,
                len(self._eigvals_ii[0]),
            )
            randn_ii = rng.standard_normal(size=shape)
            if self._ij:
                shape = (
                    len(self._eigvals_ij),
                    2,
                    number_of_snapshots,
                    len(self._eigvals_ij[0]),
                )
                randn_ij = rng.standard_normal(size=shape)
        else:
            randn_ii = randn[0]
            randn_ij = randn[1]

        u_ii, self._conditions_ii = self._solve_ii(T, number_of_snapshots, randn_ii)
        if self._ij:
            u_ij, self._conditions_ij = self._solve_ij(T, number_of_snapshots, randn_ij)
        else:
            u_ij = 0
            self._conditions_ij = None

        mass = self._dynmat.supercell.masses.reshape(-1, 1)
        u = np.array((u_ii + u_ij) / np.sqrt(mass * N), dtype="double", order="C")
        if self._max_distance:
            dists = np.linalg.norm(u, axis=2)
            dists = np.repeat(dists, 3).reshape(dists.shape + (-1,))
            self._u = np.array(
                np.where(dists < self._max_distance, u, u / dists * self._max_distance),
                dtype="double",
                order="C",
            )
        else:
            self._u = u

    @property
    def u(self):
        """Return random displacements."""
        return self._u

    @property
    def uu(self):
        """Return displacement-displacement correlation matrix."""
        return self._uu

    @property
    def uu_inv(self):
        """Return inversion of displacement-displacement correlation matrix."""
        return self._uu_inv

    @property
    def frequencies(self):
        """Setter and getter of phonon frequencies.

        Phonon frequences themselves are not stored in this instance, but are
        stored in a way of eigenvalues. The eigenvalues can be stored vai
        frequencies setter attributed.

        """
        if self._ij:
            eigvals = np.vstack((self._eigvals_ii, self._eigvals_ij))
        else:
            eigvals = self._eigvals_ii
        freqs = np.sqrt(np.abs(eigvals)) * np.sign(eigvals) * self._factor
        return np.array(freqs, dtype="double", order="C")

    @frequencies.setter
    def frequencies(self, freqs):
        eigvals = (freqs / self._factor) ** 2 * np.sign(freqs)
        if len(eigvals) != len(self._eigvals_ii) + len(self._eigvals_ij):
            raise RuntimeError("Dimension of frequencies is wrong.")

        self._eigvals_ii = eigvals[: len(self._eigvals_ii)]
        self._eigvals_ij = eigvals[len(self._eigvals_ii) :]

    @property
    def qpoints(self):
        """Return commensurate q-points where phonons are computed.."""
        N = len(self._comm_points)
        return self._comm_points[self._ii + self._ij] / float(N)

    @property
    def integrated_modes(self):
        """Return commensurate q-points where phonons are computed.."""
        if self._conditions_ij is None:
            return self._conditions_ii
        else:
            return np.vstack((self._conditions_ii, self._conditions_ij))

    @property
    def force_constants(self):
        """Return force constants."""
        return self._force_constants

    def run_d2f(self):
        """Calculate force constants from phonon eigen-solutions."""
        qpoints, eigvals, eigvecs = self._collect_eigensolutions()
        d2f = DynmatToForceConstants(
            self._dynmat.primitive,
            self._dynmat.supercell,
            use_openmp=self._dynmat.use_openmp,
        )
        d2f.commensurate_points = qpoints
        d2f.create_dynamical_matrices(eigvals, eigvecs)
        d2f.run()
        self._force_constants = d2f.force_constants

    def run_correlation_matrix(self, T):
        """Calculate displacement-displacement correlation matrix."""
        qpoints, eigvals, eigvecs = self._collect_eigensolutions()
        d2f = DynmatToForceConstants(
            self._dynmat.primitive,
            self._dynmat.supercell,
            use_openmp=self._dynmat.use_openmp,
        )
        masses = self._dynmat.supercell.masses
        d2f.commensurate_points = qpoints
        freqs = np.sqrt(np.abs(eigvals)) * self._factor
        conditions = freqs > self._cutoff_frequency
        a, _ = self._get_sigma(eigvals, T)
        a2 = a**2
        _a = np.where(conditions, a, 1)
        a2_inv = np.where(conditions, 1 / _a**2, 0)

        d2f.create_dynamical_matrices(a2_inv, eigvecs)
        d2f.run()
        self._uu_inv = np.array(d2f.force_constants, dtype="double", order="C")

        d2f.create_dynamical_matrices(a2, eigvecs)
        d2f.run()
        matrix = d2f.force_constants
        for i, m_i in enumerate(masses):
            for j, m_j in enumerate(masses):
                matrix[i, j] /= m_i * m_j
        self._uu = np.array(matrix, dtype="double", order="C")

    def treat_imaginary_modes(self, freq_from=0.01, freq_to=0.5, freq_shift=1.0):
        """Apply treatment to imaginary modes.

        This method modifies force constants to make phonon frequencies
        be real from imaginary. This treatment is expected to be finally
        forgotten after many iterations. Therefore it is unnecessary
        to be physical and can be physically dirty. If it works, it is OK,
        though good treatment may contribute to quick convergence.

        1) All frequencies at commensurate points are converted to their
        absolute values. freqs -> |freqs|.
        2) Phonon frequencies in the interval freq_from < |freqs| < freq_to
        are shifted by |freqs| + 1.

        """
        freqs = np.abs(self.frequencies)
        condition = np.logical_and(freqs > freq_from, freqs < freq_to)
        # self.frequencies is a setter attribute that sets eigenvalues.
        self.frequencies = np.where(condition, freqs + freq_shift, freqs)
        self.run_d2f()

    def _collect_eigensolutions(self):
        N = len(self._comm_points)

        qpoints = self._comm_points[self._ii] / float(N)
        eigvals = self._eigvals_ii
        eigvecs = []
        # Transform eigenvectors of D-type to those of C-type
        for q, eigvec in zip(qpoints, self._eigvecs_ii):
            Vd = np.repeat(np.exp(-2j * np.pi * np.dot(self._ppos, q)), 3)
            eigvecs.append((Vd * eigvec.T).T)

        if self._ij:
            eigvals = np.vstack((eigvals, self._eigvals_ij, self._eigvals_ij))
            eigvecs = np.vstack((eigvecs, self._eigvecs_ij, self._eigvecs_ij))
            eigvecs[-len(self._ij) :] = eigvecs[-len(self._ij) :].conj()
            qpoints = self._comm_points[self._ii + self._ij * 2] / float(N)
            qpoints[-len(self._ij) :] = -qpoints[-len(self._ij) :]

        return qpoints, eigvals, eigvecs

    def _prepare(self):
        """Calculate eigensolutions and phase factors used for random disps.

        A is a set of q-points where q = -q + G. B is a set of other q-points
        than A.

        D-type dynamical matrix is used for A to describe displacements by only
        real values, i.e., eigenvectors and phase factors.
        However C-type dynamical matrix is used for B in this implementation.

        """
        N = len(self._comm_points)
        self._eigvals_ii = []
        self._eigvecs_ii = []
        self._phase_ii = []
        self._eigvals_ij = []
        self._eigvecs_ij = []
        self._phase_ij = []

        # q in A
        for q in self._comm_points[self._ii] / float(N):
            self._dynmat.run(q)
            dm = self._C_to_D(self._dynmat.dynamical_matrix, q)
            self._phase_ii.append(
                np.cos(2 * np.pi * np.dot(self._lpos, q)).reshape(-1, 1)
            )
            eigvals, eigvecs = np.linalg.eigh(dm)
            self._eigvals_ii.append(eigvals)
            self._eigvecs_ii.append(eigvecs)

        # q in B
        if self._ij:
            for q in self._comm_points[self._ij] / float(N):
                self._dynmat.run(q)
                dm = self._dynmat.dynamical_matrix
                eigvals, eigvecs = np.linalg.eigh(dm)
                self._eigvals_ij.append(eigvals.real)
                self._eigvecs_ij.append(eigvecs)
                self._phase_ij.append(
                    np.exp(2j * np.pi * np.dot(self._spos, q)).reshape(-1, 1)
                )

    def _C_to_D(self, dm, q):
        """Transform C-type dynamical matrix to D-type."""
        V = np.repeat(np.exp(2j * np.pi * np.dot(self._ppos, q)), 3)
        dm = ((V * (V.conj() * dm).T).T).real  # C-type to D-type
        return dm

    def _setup_sampling_qpoints(self, slat, plat):
        smat = np.rint(np.dot(slat, np.linalg.inv(plat)).T).astype(int)
        self._comm_points = get_commensurate_points_in_integers(smat)
        self._ii, self._ij = categorize_commensurate_points(self._comm_points)

    def _solve_ii(
        self,
        T: float,
        number_of_snapshots: int,
        randn: np.ndarray,
    ):
        """Solve ii terms.

        randn parameter is used for the test.

        """
        natom = len(self._dynmat.supercell)
        u = np.zeros((number_of_snapshots, natom, 3), dtype="double")

        sigmas, conditions = self._get_sigma(self._eigvals_ii, T)
        for norm_dist, sigma, eigvecs, phase in zip(
            randn, sigmas, self._eigvecs_ii, self._phase_ii
        ):
            u_red = np.dot(norm_dist * sigma, eigvecs.T).reshape(
                number_of_snapshots, -1, 3
            )[:, self._s2pp, :]
            # u_red.shape = (snapshots, satoms, 3)
            # phase.shape = (satoms,)
            u += u_red * phase

        return u, conditions

    def _solve_ij(
        self,
        T: float,
        number_of_snapshots: int,
        randn: Optional[np.ndarray] = None,
    ):
        """Solve ij terms.

        randn parameter is used for the test.

        """
        natom = len(self._dynmat.supercell)
        u = np.zeros((number_of_snapshots, natom, 3), dtype="double")
        sigmas, conditions = self._get_sigma(self._eigvals_ij, T)
        for norm_dist, sigma, eigvecs, phase in zip(
            randn, sigmas, self._eigvecs_ij, self._phase_ij
        ):
            u_red = np.dot(norm_dist * sigma, eigvecs.T).reshape(
                2, number_of_snapshots, -1, 3
            )[:, :, self._s2pp, :]
            # u_red.shape = (2, snapshots, satoms, 3)
            # phase.shape = (satoms,)
            u += (u_red[0] * phase).real
            u -= (u_red[1] * phase).imag

        return u * np.sqrt(2), conditions

    def _get_sigma(self, eigvals, T):
        """Return sigma in sqrt(AMU).Angstrom unit."""
        freqs = np.sqrt(np.abs(eigvals)) * self._factor
        conditions = freqs > self._cutoff_frequency
        freqs = np.where(conditions, freqs, 1)
        if self._dist_func == "classical":
            sigma = np.sqrt(T * self._unit_conversion_classical) / freqs
        else:
            n = bose_einstein_dist(freqs, T)
            sigma = np.sqrt(self._unit_conversion / freqs * (0.5 + n))
        sigma = np.where(conditions, sigma, 0)
        return sigma, conditions



"""File I/O related routines."""

# Copyright (C) 2011 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import io
import pathlib
import sys
from collections.abc import Sequence
from typing import Optional, Union

import numpy as np
import yaml

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader

from phonopy.cui.settings import fracval
from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.dataset import get_displacements_and_forces
from phonopy.structure.symmetry import Symmetry, elaborate_borns_and_epsilon
from phonopy.utils import similarity_transformation


#
# FORCE_SETS
#
def write_FORCE_SETS(dataset, filename="FORCE_SETS"):
    """Write FORCE_SETS from dataset.

    See more detail in ``get_FORCE_SETS_lines``.

    """
    lines = get_FORCE_SETS_lines(dataset)
    with open(filename, "w") as w:
        w.write("\n".join(lines))
        w.write("\n")


def get_FORCE_SETS_lines(dataset, forces=None):
    """Generate FORCE_SETS string.

    See the format of dataset in the docstring of
    Phonopy.dataset. Optionally, sets of forces of supercells
    can be given. In this case, these forces are unnecessary to be stored
    in the dataset.

    """
    if "first_atoms" in dataset:
        return _get_FORCE_SETS_lines_type1(dataset, forces=forces)
    elif "displacements" in dataset:
        if forces is not None:
            dataset["forces"] = forces
        return _get_FORCE_SETS_lines_type2(dataset)


def _get_FORCE_SETS_lines_type1(dataset, forces=None):
    num_atom = dataset["natom"]
    displacements = dataset["first_atoms"]
    if forces is None:
        _forces = [x["forces"] for x in dataset["first_atoms"]]
    else:
        _forces = forces

    lines = []
    lines.append("%-5d" % num_atom)
    lines.append("%-5d" % len(displacements))
    for count, disp in enumerate(displacements):
        lines.append("")
        lines.append("%-5d" % (disp["number"] + 1))
        lines.append("%20.16f %20.16f %20.16f" % tuple(disp["displacement"]))
        for f in _forces[count]:
            lines.append("%15.10f %15.10f %15.10f" % tuple(f))

    return lines


def _get_FORCE_SETS_lines_type2(dataset):
    lines = []
    for displacements, forces in zip(dataset["displacements"], dataset["forces"]):
        for d, f in zip(displacements, forces):
            lines.append(("%15.8f" * 6) % (tuple(d) + tuple(f)))

    return lines


def parse_FORCE_SETS(natom=None, filename="FORCE_SETS", to_type2=False):
    """Parse FORCE_SETS from file.

    to_type2 : bool
        dataset of type2 is returned when True.

    Returns
    -------
    dataset : dict
        Displacement dataset. See Phonopy.dataset.

    """
    with open(filename, "r") as f:
        return _get_dataset(
            f,
            natom=natom,
            to_type2=to_type2,
        )


def parse_FORCE_SETS_from_strings(strings, natom=None, to_type2=False):
    """Parse FORCE_SETS from strings."""
    return _get_dataset(io.StringIO(strings), natom=natom, to_type2=to_type2)


def _get_dataset(f, natom=None, to_type2=False):
    first_line_ary = _get_line_ignore_blank(f).split()
    f.seek(0)
    if len(first_line_ary) == 1:
        if natom is None or int(first_line_ary[0]) == natom:
            dataset = _get_dataset_type1(f)
        else:
            msg = "Number of forces is not consistent with supercell setting."
            raise RuntimeError(msg)

        if to_type2:
            disps, forces = get_displacements_and_forces(dataset)
            return {"displacements": disps, "forces": forces}
        else:
            return dataset

    elif len(first_line_ary) == 6:
        return get_dataset_type2(f, natom)


def _get_dataset_type1(f):
    set_of_forces = []
    num_atom = int(_get_line_ignore_blank(f))
    num_displacements = int(_get_line_ignore_blank(f))

    for _ in range(num_displacements):
        line = _get_line_ignore_blank(f)
        atom_number = int(line)
        line = _get_line_ignore_blank(f).split()
        displacement = np.array([float(x) for x in line])
        forces_tmp = []
        for _ in range(num_atom):
            line = _get_line_ignore_blank(f).split()
            forces_tmp.append(np.array([float(x) for x in line]))
        forces_tmp = np.array(forces_tmp, dtype="double")
        forces = {
            "number": atom_number - 1,
            "displacement": displacement,
            "forces": forces_tmp,
        }
        set_of_forces.append(forces)

    dataset = {"natom": num_atom, "first_atoms": set_of_forces}

    return dataset


def get_dataset_type2(f, natom):
    """Parse type2 FORCE_SETS text and return dataset."""
    data = np.loadtxt(f, dtype="double")
    if data.shape[1] != 6 or (natom and data.shape[0] % natom != 0):
        msg = "Data shape of forces and displacements is incorrect."
        raise RuntimeError(msg)
    if natom:
        data = data.reshape(-1, natom, 6)
        displacements = data[:, :, :3]
        forces = data[:, :, 3:]
    else:
        displacements = data[:, :3]
        forces = data[:, 3:]
    dataset = {
        "displacements": np.array(displacements, dtype="double", order="C"),
        "forces": np.array(forces, dtype="double", order="C"),
    }
    return dataset


def _get_line_ignore_blank(f):
    line = f.readline().strip()
    if line == "":
        line = _get_line_ignore_blank(f)
    return line


def collect_forces(f, num_atom, hook, force_pos, word=None):
    """General function to collect forces from lines of a text file.

    Parameters
    ----------
    f :
        Text file pointer such as that returned by ``open(filename)``.
    num_atom : int
        Number of atoms in cell. Quit parsing when number of forces reaches this
        number.
    hook : str
        When this word is found at a line, parsing will start from the next line.
    force_pos : list
        Positions of force values in `line.split()`.
    word : str, optional
        Lines containing this word is only parsed. Default is None.

    Example
    -------
    The following is the abinit output.

    ...
    cartesian forces (hartree/bohr) at end:
    1     -0.00093686935947    -0.00000000000000    -0.00000000000000
    2      0.00015427277409    -0.00000000000000    -0.00000000000000
    3     -0.00000200377550    -0.00000000000000    -0.00000000000000
    4      0.00000619017547    -0.00000000000000    -0.00000000000000
    ...

    hook = "cartesian forces (eV/Angstrom)"
    force_pos = [1, 2, 3]

    """
    for line in f:
        if hook in line:
            break

    forces = []
    for line in f:
        if line.strip() == "":
            continue
        if word is not None:
            if word not in line:
                continue

        elems = line.split()
        if len(elems) > force_pos[2]:
            try:
                forces.append([float(elems[i]) for i in force_pos])
            except ValueError:
                forces = []
                break
        else:
            return False

        if len(forces) == num_atom:
            break

    return forces


def iter_collect_forces(filename, num_atom, hook, force_pos, word=None, max_iter=1000):
    """Repeat ``collect_forces`` to get the last set of forces in the file.

    Details of parameters are explained in ``collect_forces``.

    """
    with open(filename) as f:
        forces = []
        prev_forces = []

        for i in range(max_iter):  # noqa B007
            forces = collect_forces(f, num_atom, hook, force_pos, word=word)
            if not forces:
                forces = prev_forces[:]
                break
            else:
                prev_forces = forces[:]

        if i == max_iter - 1:
            sys.stderr.write("Reached to max number of iterations (%d).\n" % max_iter)

        return forces


#
# FORCE_CONSTANTS, force_constants.hdf5
#
def write_FORCE_CONSTANTS(force_constants, filename="FORCE_CONSTANTS", p2s_map=None):
    """Write force constants in text file format.

    Parameters
    ----------
    force_constants: ndarray
        Force constants
        shape=(n_satom,n_satom,3,3) or (n_patom,n_satom,3,3)
        dtype=double
    filename: str
        Filename to be saved
    p2s_map: ndarray
        Primitive atom indices in supercell index system
        dtype=intc

    """
    lines = get_FORCE_CONSTANTS_lines(force_constants, p2s_map=p2s_map)
    with open(filename, "w") as w:
        w.write("\n".join(lines))


def get_FORCE_CONSTANTS_lines(force_constants, p2s_map=None):
    """Return text in FORCE_CONSTANTS format.

    See also ``write_FORCE_CONSTANTS``.

    """
    if p2s_map is not None and len(p2s_map) == force_constants.shape[0]:
        indices = p2s_map
    else:
        indices = np.arange(force_constants.shape[0], dtype="intc")

    lines = []
    fc_shape = force_constants.shape
    lines.append("%4d %4d" % fc_shape[:2])
    for i, s_i in enumerate(indices):
        for j in range(fc_shape[1]):
            lines.append("%d %d" % (s_i + 1, j + 1))
            for vec in force_constants[i][j]:
                lines.append(("%22.15f" * 3) % tuple(vec))

    return lines


def write_force_constants_to_hdf5(
    force_constants: np.ndarray,
    filename: str = "force_constants.hdf5",
    p2s_map: Optional[np.ndarray] = None,
    physical_unit: Optional[str] = None,
    compression: Optional[Union[str, int]] = None,
):
    """Write force constants in hdf5 format.

    Parameters
    ----------
    force_constants: ndarray
        Force constants
        shape=(n_satom,n_satom,3,3) or (n_patom,n_satom,3,3)
        dtype=double
    filename: str
        Filename to be saved
    p2s_map: ndarray
        Primitive atom indices in supercell index system
        shape=(n_patom,)
        dtype=intc
    physical_unit : str, optional
        Physical unit used for force contants. Default is None.
    compression : str or int, optional
        h5py's lossless compression filters (e.g., "gzip", "lzf").
        See the detail at docstring of h5py.Group.create_dataset. Default is
        None.

    """
    try:
        import h5py
    except ImportError as exc:
        raise ModuleNotFoundError("You need to install python-h5py.") from exc

    with h5py.File(filename, "w") as w:
        w.create_dataset(
            "force_constants", data=force_constants, compression=compression
        )
        if p2s_map is not None:
            w.create_dataset("p2s_map", data=p2s_map)
        if physical_unit is not None:
            w.create_dataset("physical_unit", data=[physical_unit])


def parse_FORCE_CONSTANTS(filename="FORCE_CONSTANTS", p2s_map=None):
    """Parse FORCE_CONSTANTS.

    Parameters
    ----------
    filename : str, optional
        Filename.
    p2s_map : ndarray, optional
        Primitive.p2s_map. Supplied, this is used to check file format consistency.

    """
    with open(filename) as fcfile:
        idx1 = []

        line = fcfile.readline()
        idx = [int(x) for x in line.split()]
        if len(idx) == 1:
            idx = [idx[0], idx[0]]
        force_constants = np.zeros((idx[0], idx[1], 3, 3), dtype="double")
        for i in range(idx[0]):
            for j in range(idx[1]):
                s_i = int(fcfile.readline().split()[0]) - 1
                if s_i not in idx1:
                    idx1.append(s_i)
                tensor = []
                for _ in range(3):
                    tensor.append([float(x) for x in fcfile.readline().split()])
                force_constants[i, j] = tensor

        check_force_constants_indices(idx, idx1, p2s_map, filename)

        return force_constants


def read_force_constants_hdf5(
    filename="force_constants.hdf5", p2s_map=None, return_physical_unit=False
):
    """Parse force_constants.hdf5.

    Parameters
    ----------
    filename : str, optional
        Filename.
    p2s_map : ndarray, optional
        Primitive.p2s_map. Supplied, this is used to check file format consistency.
    return_physical_unit : bool, optional
        When True and physical_unit is in file, physical unit is returned.
        Default is False.

    """
    try:
        import h5py
    except ImportError as exc:
        raise ModuleNotFoundError("You need to install python-h5py.") from exc

    with h5py.File(filename, "r") as f:
        if "fc2" in f:
            key = "fc2"
        elif "force_constants" in f:
            key = "force_constants"
        else:
            raise RuntimeError("%s doesn't contain necessary information" % filename)

        fc = f[key][:]
        if "p2s_map" in f:
            p2s_map_in_file = f["p2s_map"][:]
            check_force_constants_indices(
                fc.shape[:2], p2s_map_in_file, p2s_map, filename
            )

        if return_physical_unit:
            if "physical_unit" in f:
                physical_unit = f["physical_unit"][0].decode("utf-8")
            else:
                physical_unit = None
            return fc, physical_unit
        else:
            return fc


def check_force_constants_indices(shape, indices, p2s_map, filename):
    """Check consistency of force constants data type."""
    if shape[0] != shape[1] and p2s_map is not None:
        if len(p2s_map) != len(indices) or (p2s_map != indices).any():
            lines = [
                f"{filename} file is inconsistent with the calculation setting. "
                "PRIMITIVE_AXIS may not be set correctly.",
                "p2s_map in primitive != p2s_map in file",
                f"{p2s_map} != {indices}",
            ]
            raise RuntimeError("\n".join(lines))


def parse_disp_yaml(filename="disp.yaml", return_cell=False):
    """Read disp.yaml or phonopy_disp.yaml.

    This method was originally made for parsing disp.yaml. Later this
    started to work for phonopy_disp.yaml, too. But now this method is not
    allowed to read phonopy_disp.yaml because of existance of PhonopyYaml
    class.

    """
    with open(filename) as f:
        new_dataset = {}
        dataset = yaml.load(f, Loader=Loader)
        if "phonopy" in dataset and "calculator" in dataset["phonopy"]:
            new_dataset["calculator"] = dataset["phonopy"]["calculator"]
        if "natom" in dataset:
            natom = dataset["natom"]
        elif "supercell" and "points" in dataset["supercell"]:
            natom = len(dataset["supercell"]["points"])
        else:
            raise RuntimeError("%s doesn't contain necessary information.")
        new_dataset["natom"] = natom
        new_first_atoms = []

        try:
            displacements = dataset["displacements"]
        except KeyError:
            raise

        if isinstance(displacements[0], dict):
            for first_atoms in displacements:
                first_atoms["atom"] -= 1
                atom1 = first_atoms["atom"]
                disp1 = first_atoms["displacement"]
                new_first_atoms.append({"number": atom1, "displacement": disp1})
            new_dataset["first_atoms"] = new_first_atoms

        if return_cell:
            cell = get_cell_from_disp_yaml(dataset)
            return new_dataset, cell
        else:
            return new_dataset


def write_disp_yaml_from_dataset(dataset, supercell, filename="disp.yaml"):
    """Write disp.yaml from dataset.

    This function is obsolete, because disp.yaml is obsolete.

    """
    displacements = [
        (d["number"],) + tuple(d["displacement"]) for d in dataset["first_atoms"]
    ]
    write_disp_yaml(displacements, supercell, filename=filename)


def write_disp_yaml(displacements: dict, supercell: PhonopyAtoms, filename="disp.yaml"):
    """Write disp.yaml from displacements.

    This function is obsolete, because disp.yaml is obsolete.

    """
    lines = []
    lines.append("natom: %4d" % len(supercell))
    lines += _get_disp_yaml_lines(displacements, supercell)
    lines.append(str(supercell))

    with open(filename, "w") as w:
        w.write("\n".join(lines))


def _get_disp_yaml_lines(displacements, supercell):
    lines = []
    lines.append("displacements:")
    for _, disp in enumerate(displacements):
        lines.append("- atom: %4d" % (disp[0] + 1))
        lines.append("  displacement:")
        lines.append("    [ %20.16f,%20.16f,%20.16f ]" % tuple(disp[1:4]))
    return lines


#
# DISP (old phonopy displacement format)
#
def parse_DISP(filename="DISP"):
    """Parse DISP file.

    This function is obsolete, because DISP is obsolete.

    """
    with open(filename) as disp:
        displacements = []
        for line in disp:
            if line.strip() != "":
                a = line.split()
                displacements.append(
                    [int(a[0]) - 1, float(a[1]), float(a[2]), float(a[3])]
                )
        return displacements


#
# Parse supercell in disp.yaml
#
def get_cell_from_disp_yaml(dataset):
    """Read cell from disp.yaml like file."""
    if "lattice" in dataset:
        lattice = dataset["lattice"]
        if "points" in dataset:
            data_key = "points"
            pos_key = "coordinates"
        elif "atoms" in dataset:
            data_key = "atoms"
            pos_key = "position"
        else:
            data_key = None
            pos_key = None

        try:
            positions = [x[pos_key] for x in dataset[data_key]]
        except KeyError as exc:
            msg = (
                '"disp.yaml" format is too old. '
                'Please re-create it as "phonopy_disp.yaml" to contain '
                "supercell crystal structure information."
            )
            raise RuntimeError(msg) from exc
        symbols = [x["symbol"] for x in dataset[data_key]]
        cell = PhonopyAtoms(cell=lattice, scaled_positions=positions, symbols=symbols)
        return cell
    else:
        return get_cell_from_disp_yaml(dataset["supercell"])


#
# QPOINTS
#
def parse_QPOINTS(filename="QPOINTS"):
    """Read QPOINTS file."""
    with open(filename, "r") as f:
        num_qpoints = int(f.readline().strip())
        qpoints = []
        for _ in range(num_qpoints):
            qpoints.append([fracval(x) for x in f.readline().strip().split()])
        return np.array(qpoints)


#
# BORN
#
def write_BORN(primitive, borns, epsilon, filename="BORN"):
    """Write BORN from NAC paramters."""
    lines = get_BORN_lines(primitive, borns, epsilon)
    with open(filename, "w") as w:
        w.write("\n".join(lines))


def get_BORN_lines(
    unitcell,
    borns,
    epsilon,
    factor=None,
    primitive_matrix=None,
    supercell_matrix=None,
    symprec=1e-5,
):
    """Generate text of BORN file."""
    borns, epsilon, atom_indices = elaborate_borns_and_epsilon(
        unitcell,
        borns,
        epsilon,
        symmetrize_tensors=True,
        primitive_matrix=primitive_matrix,
        supercell_matrix=supercell_matrix,
        symprec=symprec,
    )

    text = "# epsilon and Z* of atoms "
    text += " ".join(["%d" % n for n in atom_indices + 1])
    lines = [
        text,
    ]
    lines.append(("%13.8f " * 9) % tuple(epsilon.flatten()))
    for z in borns:
        lines.append(("%13.8f " * 9) % tuple(z.flatten()))
    return lines


def parse_BORN(primitive, symprec=1e-5, is_symmetry=True, filename="BORN"):
    """Parse BORN file.

    Parameters
    ----------
    primitive : Primitive
        Primitive cell.
    symprec : float, optional
        Symmetry tolerance. Default is 1e-5.
    is_symmetry : bool, optional
        When True, parse values are symmetrized. Default is True.
    filename : str, optional
        Filename.

    """
    with open(filename, "r") as f:
        return _parse_BORN_from_file_object(f, primitive, symprec, is_symmetry)


def parse_BORN_from_strings(
    strings, primitive, symprec=1e-5, is_symmetry=True
) -> Optional[dict]:
    """Parse BORN file text.

    See `parse_BORN` for parameters.

    """
    f = io.StringIO(strings)
    return _parse_BORN_from_file_object(f, primitive, symprec, is_symmetry)


def _parse_BORN_from_file_object(f, primitive, symprec, is_symmetry) -> Optional[dict]:
    symmetry = Symmetry(primitive, symprec=symprec, is_symmetry=is_symmetry)
    return get_born_parameters(f, primitive, symmetry)


def get_born_parameters(
    f: io.IOBase, primitive: PhonopyAtoms, prim_symmetry: Symmetry
) -> Optional[dict]:
    """Parse BORN file text.

    Parameters
    ----------
    f :
        File pointer of BORN file.
    primitive : Primitive
        Primitive cell.
    prim_symmetry : Symmetry
        Symmetry of primitive cell.

    """
    line_arr = f.readline().split()
    if len(line_arr) < 1:
        print("BORN file format of line 1 is incorrect")
        return None

    factor = None
    G_cutoff = None
    Lambda = None

    if len(line_arr) > 0:
        try:
            factor = float(line_arr[0])
        except (ValueError, TypeError):
            factor = None
    if len(line_arr) > 1:
        try:
            G_cutoff = float(line_arr[1])
        except (ValueError, TypeError):
            G_cutoff = None
    if len(line_arr) > 2:
        try:
            Lambda = float(line_arr[2])
        except (ValueError, TypeError):
            Lambda = None

    # Read dielectric constant
    line = f.readline().split()
    if not len(line) == 9:
        print("BORN file format of line 2 is incorrect")
        return None
    dielectric = np.reshape([float(x) for x in line], (3, 3))

    # Read Born effective charge
    independent_atoms = prim_symmetry.get_independent_atoms()
    borns = np.zeros((len(primitive), 3, 3), dtype="double", order="C")

    for i in independent_atoms:
        line = f.readline().split()
        if len(line) == 0:
            print("Number of lines for Born effect charge is not enough.")
            return None
        if not len(line) == 9:
            print("BORN file format of line %d is incorrect" % (i + 3))
            return None
        borns[i] = np.reshape([float(x) for x in line], (3, 3))

    # Check that the number of atoms in the BORN file was correct
    line = f.readline().split()
    if len(line) > 0:
        print(
            "Too many atoms in the BORN file (it should only contain "
            "symmetry-independent atoms)"
        )
        return None

    _expand_borns(borns, primitive, prim_symmetry)
    non_anal = {"born": borns, "dielectric": dielectric}
    if factor is not None:
        non_anal["factor"] = factor
    if G_cutoff is not None:
        non_anal["G_cutoff"] = G_cutoff
    if Lambda is not None:
        non_anal["Lambda"] = Lambda

    return non_anal


def _expand_borns(borns, primitive: PhonopyAtoms, prim_symmetry: Symmetry):
    # Expand Born effective charges to all atoms in the primitive cell
    rotations = prim_symmetry.symmetry_operations["rotations"]
    map_operations = prim_symmetry.get_map_operations()
    map_atoms = prim_symmetry.get_map_atoms()

    for i in range(len(primitive)):
        # R_cart = L R L^-1
        rot_cartesian = similarity_transformation(
            primitive.cell.T, rotations[map_operations[i]]
        )
        # R_cart^T B R_cart^-T (inverse rotation is required to transform)
        borns[i] = similarity_transformation(rot_cartesian.T, borns[map_atoms[i]])


#
# phonopy.yaml
#
def is_file_phonopy_yaml(
    filename, keyword: Optional[str] = None, yaml_dict_keys: Optional[Sequence] = None
):
    """Check whether the file is phonopy.yaml like file or not.

    Parameters
    ----------
    filename : str
        Filename.
    keyword : str, optional
        Deprecated. Dummy parameter. Default is None.
    yaml_dict_keys : list-like, optional
        List of necessary keys in the yaml file. Default is None.

    """

    def keys_exist(yaml_data: dict, keys: Sequence[str]) -> bool:
        for key in keys:
            if key not in yaml_data:
                return False
        return True

    myio = get_io_module_to_decompress(filename)
    with myio.open(filename, "r") as f:
        try:
            data = yaml.load(f, Loader=Loader)
            if data is None:
                return False
            if yaml_dict_keys is None:
                keys = ("supercell_matrix", "unit_cell")
            else:
                keys = yaml_dict_keys
            return keys_exist(data, keys)
        except yaml.YAMLError:
            return False


#
# e-v.dat, thermal_properties.yaml
#
def read_thermal_properties_yaml(filenames):
    """Read thermal_properties.yaml."""
    thermal_properties = []
    num_modes = []
    num_integrated_modes = []
    for filename in filenames:
        with open(filename) as f:
            tp_yaml = yaml.load(f, Loader=Loader)
            thermal_properties.append(tp_yaml["thermal_properties"])
            if "num_modes" in tp_yaml and "num_integrated_modes" in tp_yaml:
                num_modes.append(tp_yaml["num_modes"])
                num_integrated_modes.append(tp_yaml["num_integrated_modes"])

    temperatures = [v["temperature"] for v in thermal_properties[0]]
    temp = []
    cv = []
    entropy = []
    fe_phonon = []
    for _, tp in enumerate(thermal_properties):
        temp.append([v["temperature"] for v in tp])
        if not np.allclose(temperatures, temp):
            msg = [
                "",
            ]
            msg.append("Check your input files")
            msg.append("Disagreement of temperature range or step")
            for t, fname in zip(temp, filenames):
                msg.append(
                    "%s: Range [ %d, %d ], Step %f"
                    % (fname, int(t[0]), int(t[-1]), t[1] - t[0])
                )
            msg.append("")
            msg.append("Stop phonopy-qha")
            raise RuntimeError(msg)
        cv.append([v["heat_capacity"] for v in tp])
        entropy.append([v["entropy"] for v in tp])
        fe_phonon.append([v["free_energy"] for v in tp])

    # shape=(temperatures, volumes)
    cv = np.array(cv).T
    entropy = np.array(entropy).T
    fe_phonon = np.array(fe_phonon).T

    return (temperatures, cv, entropy, fe_phonon, num_modes, num_integrated_modes)


def read_v_e(filename):
    """Read v-e.dat file."""
    data = _parse_QHA_data(filename)
    if data.shape[1] != 2:
        msg = "File format of %s is incorrect for reading e-v data." % filename
        raise RuntimeError(msg)
    volumes, electronic_energies = data.T
    return volumes, electronic_energies


def read_efe(filename):
    """Read fe-v.dat (efe) file."""
    data = _parse_QHA_data(filename)
    temperatures = data[:, 0]
    free_energies = data[:, 1:]
    return temperatures, free_energies


def _parse_QHA_data(filename):
    data = []
    with open(filename) as f:
        for line in f:
            if line.strip() == "" or line.strip()[0] == "#":
                continue
            if "#" in line:
                data.append([float(x) for x in line.split("#")[0].split()])
            else:
                data.append([float(x) for x in line.split()])
        return np.array(data)


def get_io_module_to_decompress(filename):
    """Return io-module to decompress file.

    Filename extensions of lzma, xz, gzip, bz2 are supported.

    It is supported to use it like `returned_module.open(filename)`.

    """
    ext = pathlib.Path(filename).suffix
    if ext == ".xz" or ext == ".lzma":
        import lzma

        return lzma
    elif ext == ".gz":
        import gzip

        return gzip
    elif ext == ".bz2":
        import bz2

        return bz2
    else:
        import io

        return io


def get_supported_file_extensions_for_compression():
    """Return file extensions for compression.

    This function must be coupled with `get_io_module_to_decompress`.

    """
    return "", ".xz", ".lzma", ".gz", ".bz2"



# Copyright (C) 2012 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import sys

import numpy as np

try:
    import yaml
except ImportError:
    print("You need to install python-yaml.")
    sys.exit(1)

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader


def print_band(data):
    """Print band structure in gnuplot style."""
    distance = 0
    text = "#%-9s %-10s %-10s    " % ("    q_a", "    q_b", "    q_c")
    text += "%-12s    " % "   Distance"
    text += "%-12s %-12s     ..." % ("   Gruneisen", "   Frequency")
    lines = [
        text,
    ]

    for path in data["path"]:
        for q in path["phonon"]:
            text = "%10.7f %10.7f %10.7f    " % (tuple(q["q-position"]))
            text += "%12.7f    " % (q["distance"] + distance)
            for band in q["band"]:
                text += "%12.7f %12.7f  " % (band["gruneisen"], band["frequency"])
            lines.append(text)
        distance += path["phonon"][-1]["distance"]

    print("\n".join(lines))


def print_mesh(data):
    """Print mesh information."""
    text = "#%-9s %-10s %-10s  " % ("    q_a", "    q_b", "    q_c")
    text += "%-12s  " % "Multiplicity"
    text += "%-12s %-12s     ..." % ("   Gruneisen", "   Frequency")
    lines = [
        text,
    ]

    for qpt in data["phonon"]:
        text = "%10.7f %10.7f %10.7f  " % (tuple(qpt["q-position"]))
        text += "%12d  " % qpt["multiplicity"]
        for band in qpt["band"]:
            text += "%12.7f %12.7f  " % (band["gruneisen"], band["frequency"])
        lines.append(text)

    print("\n".join(lines))


def plot_band(data, is_fg, g_max, g_min):
    """Plot band structure."""
    import matplotlib.pyplot as plt

    d = []
    g = []
    f = []
    distance = 0.0

    for path in data["path"]:
        for q in path["phonon"]:
            d.append(q["distance"] + distance)
            g.append([band["gruneisen"] for band in q["band"]])
            f.append([band["frequency"] for band in q["band"]])
        distance += path["phonon"][-1]["distance"]

    if is_fg:
        fig, (ax1, ax2, ax3) = plt.subplots(3, 1)
        ax1.plot(d, g, "-")
        ax2.plot(d, np.array(g) * np.array(f), "-")
        ax3.plot(d, f, "-")
    else:
        fig, (ax1, ax2) = plt.subplots(2, 1)
        ax1.plot(d, g, "-")
        ax2.plot(d, f, "-")

    if g_max is not None:
        ax1.set_ylim(ymax=g_max)
    if g_min is not None:
        ax1.set_ylim(ymin=g_min)

    return plt


def plot_yaml_mesh(data, is_fg, cutoff_frequency=None):
    """Plot band structure using data from yaml."""
    x = []
    y = []
    for qpt in data["phonon"]:
        x.append([band["frequency"] for band in qpt["band"]])
        y.append([band["gruneisen"] for band in qpt["band"]])

    return plot_mesh(x, y, is_fg, cutoff_frequency=cutoff_frequency)


def plot_hdf5_mesh(data, is_fg, cutoff_frequency=None):
    """Plot band structure using data from hdf5."""
    x = data["frequency"]
    y = data["gruneisen"][:]
    return plot_mesh(x, y, is_fg, cutoff_frequency=cutoff_frequency)


def plot_mesh(x, y, is_fg, cutoff_frequency=None):
    """Plot mesh data."""
    import matplotlib.pyplot as plt

    for g, freqs in zip(np.transpose(y), np.transpose(x)):
        if cutoff_frequency:
            g = np.extract(freqs > cutoff_frequency, g)
            freqs = np.extract(freqs > cutoff_frequency, freqs)

        if is_fg:
            plt.plot(freqs, np.array(freqs) * np.array(g), "o")
        else:
            plt.plot(freqs, g, "o")

    return plt


def get_options():
    """Parse command-line options."""
    import argparse

    parser = argparse.ArgumentParser(
        description="Phonopy gruneisenplot command-line-tool"
    )
    parser.set_defaults(
        f_max=None,
        f_min=None,
        g_max=None,
        g_min=None,
        is_hdf5=False,
        is_fg=False,
        is_gnuplot=False,
        cutoff_frequency=None,
        output_filename=None,
        save_graph=None,
        title=None,
    )
    parser.add_argument(
        "--gnuplot",
        dest="is_gnuplot",
        action="store_true",
        help="Output in gnuplot data style",
    )
    parser.add_argument(
        "-o", "--output", dest="output_filename", help="Output filename of PDF plot"
    )
    parser.add_argument(
        "-s",
        "--save",
        dest="save_graph",
        action="store_true",
        help="Save plot data in pdf",
    )
    parser.add_argument(
        "--fg", dest="is_fg", action="store_true", help="Plot omega x gamma"
    )
    parser.add_argument("-t", "--title", dest="title", help="Title of plot")
    parser.add_argument(
        "--cutoff",
        dest="cutoff_frequency",
        type=float,
        help="Plot above this cutoff frequency for mesh sampling mode.",
    )
    parser.add_argument(
        "--fmax", dest="f_max", type=float, help="Maximum frequency plotted"
    )
    parser.add_argument(
        "--fmin", dest="f_min", type=float, help="Minimum frequency plotted"
    )
    parser.add_argument(
        "--gmax", dest="g_max", type=float, help="Maximum Gruneisen params plotted"
    )
    parser.add_argument(
        "--gmin", dest="g_min", type=float, help="Minimum Gruneisen params plotted"
    )
    parser.add_argument(
        "--hdf5", dest="is_hdf5", action="store_true", help="Use hdf5 to read results"
    )
    parser.add_argument(
        "filename",
        nargs="*",
        help="Filename of phonopy-gruneisen result (gruneisen.yaml)",
    )
    args = parser.parse_args()
    return args


def run():
    """Run phonopy-gruneisenplot."""
    args = get_options()

    if args.is_hdf5:
        try:
            import h5py
        except ImportError:
            print("You need to install python-h5py.")
            sys.exit(1)

        if len(args.filename) == 0:
            filename = "gruneisen.hdf5"
        else:
            filename = args.filename[0]
        with h5py.File(filename, "r") as f:
            data = {key: f[key][:] for key in list(f)}
    else:
        if len(args.filename) == 0:
            filename = "gruneisen.yaml"
        else:
            filename = args.filename[0]
        with open(filename) as f:
            data = yaml.load(f.read(), Loader=Loader)

    if args.save_graph:
        import matplotlib as mpl

        mpl.use("Agg")

    if args.output_filename:
        pdffile = args.output_filename
    else:
        pdffile = "gruneisenplot.pdf"

    if "path" in data:
        if args.is_gnuplot:
            print_band(data)
        else:
            plt = plot_band(data, args.is_fg, args.g_max, args.g_min)
            if args.title is not None:
                plt.subtitle(args.title)
            if args.save_graph:
                plt.savefig(pdffile)
            else:
                plt.show()

    if "mesh" in data:
        if args.is_hdf5:
            plt = plot_hdf5_mesh(
                data, args.is_fg, cutoff_frequency=args.cutoff_frequency
            )
        else:
            if args.is_gnuplot:
                print_mesh(data)
            else:
                plt = plot_yaml_mesh(
                    data, args.is_fg, cutoff_frequency=args.cutoff_frequency
                )

        if args.title is not None:
            plt.subtitle(args.title)

        if args.f_max is not None:
            plt.xlim(xmax=args.f_max)
        if args.f_min is not None:
            plt.xlim(xmin=args.f_min)
        if args.g_max is not None:
            plt.ylim(ymax=args.g_max)
        if args.g_min is not None:
            plt.ylim(ymin=args.g_min)

        if args.save_graph:
            plt.savefig(pdffile)
        else:
            plt.show()



"""Tests of PhonopyYaml."""

import io
from pathlib import Path

import numpy as np
import yaml

import phonopy
from phonopy import Phonopy
from phonopy.interface.phonopy_yaml import (
    PhonopyYaml,
    PhonopyYamlLoader,
    load_phonopy_yaml,
    read_cell_yaml,
    read_phonopy_yaml,
)
from phonopy.interface.vasp import read_vasp
from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.cells import get_primitive
from phonopy.structure.dataset import get_displacements_and_forces

cwd = Path(__file__).parent


def test_read_poscar_yaml(helper_methods):
    """Test to parse PhonopyAtoms.__str__ output."""
    filename = cwd / "NaCl-vasp.yaml"
    cell = _get_unitcell(filename)

    _compare_NaCl_convcell(cell, helper_methods.compare_cells_with_order)


def test_read_phonopy_yaml(helper_methods):
    """Test to parse phonopy.yaml like file."""
    filename = cwd / "phonopy.yaml"
    cell = read_phonopy_yaml(filename).unitcell
    _compare_NaCl_convcell(cell, helper_methods.compare_cells_with_order)


def test_read_phonopy_yaml_with_stream(helper_methods):
    """Test to parse phonopy.yaml like file stream."""
    filename = cwd / "phonopy.yaml"
    with open(filename) as fp:
        cell = read_phonopy_yaml(fp).unitcell
        _compare_NaCl_convcell(cell, helper_methods.compare_cells_with_order)


def test_PhonopyYaml_read(helper_methods):
    """Test to parse phonopy.yaml like file using PhonopyYaml.read()."""
    filename = cwd / "phonopy.yaml"
    cell = _get_unitcell(filename)
    _compare_NaCl_convcell(cell, helper_methods.compare_cells_with_order)


def test_PhonopyYaml_read_with_stream(helper_methods):
    """Test to parse phonopy.yaml like file stream using PhonopyYaml.read()."""
    filename = cwd / "phonopy.yaml"
    with open(filename) as fp:
        cell = _get_unitcell(fp)
        _compare_NaCl_convcell(cell, helper_methods.compare_cells_with_order)


def test_read_cell_yaml(helper_methods):
    """Test to parse phonopy_symcells.yaml like file."""
    filename = cwd / "phonopy_symcells_NaCl.yaml"
    cell = read_cell_yaml(filename)
    _compare_NaCl_convcell(cell, helper_methods.compare_cells)

    pcell = read_cell_yaml(filename, cell_type="primitive")
    helper_methods.compare_cells(pcell, get_primitive(cell, "F"))


def test_read_cell_yaml_with_stream(helper_methods):
    """Test to parse phonopy_symcells.yaml like file."""
    filename = cwd / "phonopy_symcells_NaCl.yaml"
    with open(filename) as fp:
        cell = _get_unitcell(fp)
        _compare_NaCl_convcell(cell, helper_methods.compare_cells)
        fp.seek(0)
        pcell = read_cell_yaml(fp, cell_type="primitive")
        helper_methods.compare_cells(pcell, get_primitive(cell, "F"))


def test_write_phonopy_yaml(ph_nacl_nofcsym: Phonopy, helper_methods):
    """Test PhonopyYaml.set_phonon_info, __str__, yaml_data, parse."""
    phonon = ph_nacl_nofcsym
    phpy_yaml = PhonopyYaml(calculator="vasp")
    phpy_yaml.set_phonon_info(phonon)
    phpy_yaml_test = PhonopyYaml()
    phpy_yaml_test._data = load_phonopy_yaml(
        yaml.safe_load(io.StringIO(str(phpy_yaml))), calculator=phpy_yaml.calculator
    )
    helper_methods.compare_cells_with_order(
        phpy_yaml.primitive, phpy_yaml_test.primitive
    )
    helper_methods.compare_cells_with_order(phpy_yaml.unitcell, phpy_yaml_test.unitcell)
    helper_methods.compare_cells_with_order(
        phpy_yaml.supercell, phpy_yaml_test.supercell
    )
    assert phpy_yaml.version == phpy_yaml_test.version
    np.testing.assert_allclose(
        phpy_yaml.supercell_matrix, phpy_yaml_test.supercell_matrix, atol=1e-8
    )
    np.testing.assert_allclose(
        phpy_yaml.primitive_matrix, phpy_yaml_test.primitive_matrix, atol=1e-8
    )


def test_write_phonopy_yaml_extra(ph_nacl_nofcsym: Phonopy):
    """Test PhonopyYaml.set_phonon_info, __str__, yaml_data, parse.

    settings parameter controls amount of yaml output. In this test,
    more data than the default are dumped and those are tested.

    """
    phonon = ph_nacl_nofcsym
    settings = {
        "force_sets": True,
        "displacements": True,
        "force_constants": True,
        "born_effective_charge": True,
        "dielectric_constant": True,
    }
    phpy_yaml = PhonopyYaml(calculator="vasp", settings=settings)
    phpy_yaml.set_phonon_info(phonon)
    phpy_yaml_test = PhonopyYaml()
    phpy_yaml_test._data = load_phonopy_yaml(
        yaml.safe_load(io.StringIO(str(phpy_yaml))), calculator=phpy_yaml.calculator
    )
    np.testing.assert_allclose(
        phpy_yaml.force_constants, phpy_yaml_test.force_constants, atol=1e-8
    )
    np.testing.assert_allclose(
        phpy_yaml.nac_params["born"], phpy_yaml_test.nac_params["born"], atol=1e-8
    )
    np.testing.assert_allclose(
        phpy_yaml.nac_params["dielectric"],
        phpy_yaml_test.nac_params["dielectric"],
        atol=1e-8,
    )
    np.testing.assert_allclose(
        phpy_yaml.nac_params["factor"],
        phpy_yaml_test.nac_params["factor"],
        atol=1e-8,
    )
    disps, forces = get_displacements_and_forces(phpy_yaml.dataset)
    disps_test, forces_test = get_displacements_and_forces(phpy_yaml_test.dataset)
    np.testing.assert_allclose(forces, forces_test, atol=1e-8)
    np.testing.assert_allclose(disps, disps_test, atol=1e-8)


def test_load_nac_yaml():
    """Test to read NAC params using PhonopyYamlLoader."""
    pyl = PhonopyYamlLoader(yaml.safe_load(open(cwd / "nac.yaml"))).parse()
    assert pyl.data.nac_params
    for key in (
        "dielectric",
        "born",
        "factor",
        "method",
    ):
        assert key in pyl.data.nac_params
    assert pyl.data.nac_params["dielectric"].shape == (3, 3)
    assert pyl.data.nac_params["born"].shape == (2, 3, 3)
    assert isinstance(pyl.data.nac_params["factor"], float)
    assert isinstance(pyl.data.nac_params["method"], str)


def test_phonopy_yaml_extended_symbol(nacl_unitcell_order1: PhonopyAtoms):
    """Test of PhonopyYaml with extended symbol."""
    unitcell = nacl_unitcell_order1
    symbols = unitcell.symbols
    symbols[-1] = "Cl1"
    cell = PhonopyAtoms(
        cell=unitcell.cell,
        symbols=symbols,
        scaled_positions=unitcell.scaled_positions,
        masses=unitcell.masses,
    )
    ph = Phonopy(cell, supercell_matrix=[2, 2, 2])
    assert ph.primitive.symbols[:4] == ["Na"] * 4
    assert ph.primitive.symbols[4:7] == ["Cl"] * 3
    assert ph.primitive.symbols[-1] == "Cl1"
    assert ph.unitcell.symbols[:4] == ["Na"] * 4
    assert ph.unitcell.symbols[4:7] == ["Cl"] * 3
    assert ph.unitcell.symbols[-1] == "Cl1"
    assert ph.supercell.symbols[:32] == ["Na"] * 32
    assert ph.supercell.symbols[-32:-8] == ["Cl"] * 24
    assert ph.supercell.symbols[-8:] == ["Cl1"] * 8

    ph_load = phonopy.load(io.StringIO(str(PhonopyYaml().set_phonon_info(ph))))
    assert ph_load.primitive.symbols[:4] == ["Na"] * 4
    assert ph_load.primitive.symbols[4:7] == ["Cl"] * 3
    assert ph_load.primitive.symbols[-1] == "Cl1"
    assert ph_load.unitcell.symbols[:4] == ["Na"] * 4
    assert ph_load.unitcell.symbols[4:7] == ["Cl"] * 3
    assert ph_load.unitcell.symbols[-1] == "Cl1"
    assert ph_load.supercell.symbols[:32] == ["Na"] * 32
    assert ph_load.supercell.symbols[-32:-8] == ["Cl"] * 24
    assert ph_load.supercell.symbols[-8:] == ["Cl1"] * 8


def _compare_NaCl_convcell(cell, compare_cells):
    cell_ref = read_vasp(cwd / ".." / "POSCAR_NaCl")
    compare_cells(cell, cell_ref)


def _get_unitcell(filename):
    phpy_yaml = PhonopyYaml().read(filename)
    return phpy_yaml.unitcell



"""Tests for velocity calculation from MD data."""

import os

import numpy as np

from phonopy.interface.vasp import read_XDATCAR
from phonopy.spectrum.velocity import Velocity

data_dir = os.path.dirname(os.path.abspath(__file__))


def test_Velocity():
    """Test Velocity class."""
    lattice, positions = read_XDATCAR(os.path.join(data_dir, "XDATCAR"))
    v = Velocity(positions=positions, lattice=lattice, timestep=2)
    v.run()
    velocity = v.get_velocities()
    velocity_cmp = np.loadtxt(os.path.join(data_dir, "velocities.dat"))
    assert (np.abs(velocity.ravel() - velocity_cmp.ravel()) < 1e-1).all()


def _show(velocity):
    print(velocity)


def _write(velocity):
    np.savetxt(os.path.join(data_dir, "velocities.dat"), velocity.reshape((-1, 3)))



"""Conventinal and generalized regular grid routines."""

# Copyright (C) 2014 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import warnings

import numpy as np
from spglib import (
    get_pointgroup,
    get_stabilized_reciprocal_mesh,
    get_symmetry_dataset,
    relocate_BZ_grid_address,
)

from phonopy.structure.brillouin_zone import get_qpoints_in_Brillouin_zone
from phonopy.structure.cells import (
    determinant,
    estimate_supercell_matrix,
    estimate_supercell_matrix_from_pointgroup,
    get_cell_parameters,
    get_primitive_matrix_by_centring,
)
from phonopy.structure.snf import SNF3x3
from phonopy.structure.symmetry import (
    collect_unique_rotations,
    get_lattice_vector_equivalence,
    get_pointgroup_operations,
)
from phonopy.utils import get_dot_access_dataset, similarity_transformation
from phonopy.version import __version__


def length2mesh(length, lattice, rotations=None):
    """Convert length to mesh for q-point sampling.

    This conversion for each reciprocal axis follows VASP convention by
        N = max(1, int(l * |a|^* + 0.5))
    'int' means rounding down, not rounding to nearest integer.

    Parameters
    ----------
    length : float
        Length having the unit of direct space length.
    lattice : array_like
        Basis vectors of primitive cell in row vectors.
        dtype='double', shape=(3, 3)
    rotations: array_like, optional
        Rotation matrices in real space. When given, mesh numbers that are
        symmetrically reasonable are returned. Default is None.
        dtype='int64', shape=(rotations, 3, 3)

    Returns
    -------
    array_like
        dtype=int, shape=(3,)

    """
    rec_lattice = np.linalg.inv(lattice)
    rec_lat_lengths = get_cell_parameters(rec_lattice.T)
    mesh_numbers = np.rint(rec_lat_lengths * length).astype(int)

    if rotations is not None:
        reclat_equiv = get_lattice_vector_equivalence(
            [r.T for r in np.array(rotations)]
        )
        m = mesh_numbers
        mesh_equiv = [m[1] == m[2], m[2] == m[0], m[0] == m[1]]
        # Follow symmetry when distorted, and align to larger one.
        for i, pair in enumerate(([1, 2], [2, 0], [0, 1])):
            if reclat_equiv[i] and not mesh_equiv[i]:
                m[pair] = max(m[pair])

    return np.maximum(mesh_numbers, [1, 1, 1])


def get_qpoints(
    mesh_numbers,
    reciprocal_lattice,  # column vectors
    q_mesh_shift=None,  # Monkhorst-Pack style grid shift
    is_gamma_center=True,
    is_time_reversal=True,
    fit_in_BZ=True,
    rotations=None,  # Point group operations in real space
    is_mesh_symmetry=True,
):
    """Return q-points and weights on a mesh sampling grid.

    See attributes of GriePoints, qpoints and wegiths.

    """
    gp = GridPoints(
        mesh_numbers,
        reciprocal_lattice,
        q_mesh_shift=q_mesh_shift,
        is_gamma_center=is_gamma_center,
        is_time_reversal=is_time_reversal,
        fit_in_BZ=fit_in_BZ,
        rotations=rotations,
        is_mesh_symmetry=is_mesh_symmetry,
    )

    return gp.qpoints, gp.weights


def extract_ir_grid_points(
    grid_mapping_table: np.ndarray,
) -> tuple[np.ndarray, np.ndarray]:
    """Return ir-grid points and weights in grid index mapping table."""
    dtype = grid_mapping_table.dtype
    ir_grid_points = np.array(np.unique(grid_mapping_table), dtype=dtype)
    weights = np.zeros_like(grid_mapping_table)
    for gp in grid_mapping_table:
        weights[gp] += 1
    ir_weights = np.array(weights[ir_grid_points], dtype=dtype)

    return ir_grid_points, ir_weights


class GridPoints:
    """Class to generate irreducible grid points on uniform mesh grids.

    Attributes
    ----------
    mesh_numbers: ndarray
       Mesh numbers along a, b, c axes.
       dtype='intc'
       shape=(3,)
    reciprocal_lattice: ndarray
        Basis vectors in reciprocal space. a*, b*, c* are given in column
        vectors.
        dtype='double'
        shape=(3, 3)
    qpoints: ndarray
       q-points in reduced coordinates of reciprocal lattice
       dtype='double'
       shape=(ir-grid points, 3)
    weights: ndarray
       Geometric q-point weights. Its sum is the number of grid points.
       dtype='int64'
       shape=(ir-grid points,)
    grid_address: ndarray
       Addresses of all grid points represented by integers.
       dtype='intc'
       shape=(prod(mesh_numbers), 3)
    ir_grid_points: ndarray
        Indices of irreducible grid points in grid_address.
        dtype='int64', shape=(ir-grid points,)
    grid_mapping_table: ndarray
        Index mapping table from all grid points to ir-grid points.
        dtype='int64', shape=(prod(mesh_numbers),)

    """

    def __init__(
        self,
        mesh_numbers,
        reciprocal_lattice,
        q_mesh_shift=None,  # Monkhorst-Pack style grid shift
        is_gamma_center=True,
        is_time_reversal=True,
        fit_in_BZ=True,
        rotations=None,  # Point group operations in real space
        is_mesh_symmetry=True,
    ):  # Except for time reversal symmetry
        """Init method.

        Note
        ----
        Uniform mesh grids are made according to Monkhorst-Pack scheme, i.e.,
        for odd (even) numbers, centre are (are not) sampled. The Gamma-centre
        sampling is supported by ``is_gamma_center=True``.

        Parameters
        ----------
        mesh_numbers: array_like
            Mesh numbers along a, b, c axes.
            dtype='int64'
            shape=(3, )
        reciprocal_lattice: array_like
            Basis vectors in reciprocal space. a*, b*, c* are given in column
            vectors.
            dtype='double'
            shape=(3, 3)
        q_mesh_shift: array_like, optional, default None (no shift)
            Mesh shifts along a*, b*, c* axes with respect to neighboring grid
            points from the original mesh (Monkhorst-Pack or Gamma center).
            0.5 gives half grid shift. Normally 0 or 0.5 is given.
            Otherwise q-points symmetry search is not performed.
            dtype='double'
            shape=(3, )
        is_gamma_center: bool, default False
            Uniform mesh grids are generated centring at Gamma point but not
            the Monkhorst-Pack scheme.
        is_time_reversal: bool, optional, default True
            Time reversal symmetry is considered in symmetry search. By this,
            inversion symmetry is always included.
        fit_in_BZ: bool, optional, default True
        rotations: array_like, default None (only unitary operation)
            Rotation matrices in direct space. For each rotation matrix R,
            a point in crystallographic coordinates, x, is sent as x' = Rx.
            dtype='int64'
            shape=(rotations, 3, 3)
        is_mesh_symmetry: bool, optional, default True
            Wheather symmetry search is done or not.

        """
        self._mesh = np.array(mesh_numbers, dtype="intc")
        self._rec_lat = np.array(reciprocal_lattice, dtype="double", order="C")
        self._is_shift = self._shift2boolean(
            q_mesh_shift, is_gamma_center=is_gamma_center
        )
        self._is_time_reversal = is_time_reversal
        self._fit_in_BZ = fit_in_BZ
        self._rotations = rotations
        self._is_mesh_symmetry = is_mesh_symmetry

        self._ir_qpoints = None
        self._grid_address = None
        self._ir_grid_points = None
        self._ir_weights = None
        self._grid_mapping_table = None

        if self._is_shift is None:
            self._is_mesh_symmetry = False
            self._is_shift = self._shift2boolean(None)
            self._set_grid_points()
            self._ir_qpoints += q_mesh_shift / self._mesh
            self._fit_qpoints_in_BZ()
        else:  # zero or half shift
            self._set_grid_points()

    @property
    def mesh_numbers(self):
        """Return mesh numbers."""
        return self._mesh

    @property
    def reciprocal_lattice(self):
        """Return reciprocal lattice in column vectors."""
        return self._rec_lat

    @property
    def grid_address(self):
        """Return all grid point addresses."""
        return self._grid_address

    def get_grid_address(self):
        """Return all grid point addresses."""
        warnings.warn(
            "GridPoints.get_grid_address() is deprecated."
            "Use GridPoints.grid_address attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.grid_address

    @property
    def ir_grid_points(self):
        """Return ir-grid point indices."""
        return self._ir_grid_points

    def get_ir_grid_points(self):
        """Return ir-grid point indices."""
        warnings.warn(
            "GridPoints.get_ir_grid_points() is deprecated."
            "Use GridPoints.ir_grid_points attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.ir_grid_points

    @property
    def qpoints(self):
        """Return irreducible q-points."""
        return self._ir_qpoints

    def get_ir_qpoints(self):
        """Return irreducible q-points."""
        warnings.warn(
            "GridPoints.get_ir_qpoints() is deprecated."
            "Use GridPoints.qpoints attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.qpoints

    @property
    def weights(self):
        """Return weights of ir-grid points."""
        return self._ir_weights

    def get_ir_grid_weights(self):
        """Return weights of ir-grid points."""
        warnings.warn(
            "GridPoints.get_ir_grid_weights() is deprecated."
            "Use GridPoints.weights attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.weights

    @property
    def grid_mapping_table(self):
        """Return grid index mapping table."""
        return self._grid_mapping_table

    def get_grid_mapping_table(self):
        """Return grid index mapping table."""
        warnings.warn(
            "GridPoints.get_grid_mapping_table() is deprecated."
            "Use GridPoints.grid_mapping_table attribute.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.grid_mapping_table

    def _set_grid_points(self):
        if self._is_mesh_symmetry and self._has_mesh_symmetry():
            self._set_ir_qpoints(
                self._rotations, is_time_reversal=self._is_time_reversal
            )
        else:
            self._set_ir_qpoints(
                [np.eye(3, dtype="intc")], is_time_reversal=self._is_time_reversal
            )

    def _shift2boolean(self, q_mesh_shift, is_gamma_center=False, tolerance=1e-5):
        """Return bools of with or without half-shifts.

        Parameters
        ----------
        tolerance : float
            This is used to judge zero/half gird shift.

        """
        if q_mesh_shift is None:
            shift = np.zeros(3, dtype="double")
        else:
            shift = np.array(q_mesh_shift, dtype="double")

        diffby2 = np.abs(shift * 2 - np.rint(shift * 2))

        if (diffby2 < 0.01).all():  # zero or half shift
            diff = np.abs(shift - np.rint(shift))
            if is_gamma_center:
                is_shift = list(diff > 0.1)
            else:  # Monkhorst-pack
                is_shift = list(np.logical_xor((diff > 0.1), (self._mesh % 2 == 0)) * 1)
        else:
            is_shift = None

        return is_shift

    def _has_mesh_symmetry(self):
        if self._rotations is None:
            return False
        m = self._mesh
        mesh_equiv = [m[1] == m[2], m[2] == m[0], m[0] == m[1]]
        lattice_equiv = get_lattice_vector_equivalence([r.T for r in self._rotations])
        return np.extract(lattice_equiv, mesh_equiv).all()

    def _fit_qpoints_in_BZ(self):
        qpolongset_in_BZ = get_qpoints_in_Brillouin_zone(
            self._rec_lat, self._ir_qpoints
        )
        qpoints_in_BZ = np.array(
            [q_set[0] for q_set in qpolongset_in_BZ], dtype="double", order="C"
        )
        self._ir_qpoints = qpoints_in_BZ

    def _set_ir_qpoints(self, rotations, is_time_reversal=True):
        grid_mapping_table, grid_address = get_stabilized_reciprocal_mesh(
            self._mesh,
            rotations,
            is_shift=self._is_shift,
            is_time_reversal=is_time_reversal,
            is_dense=True,
        )
        # uintp to long
        grid_mapping_table = np.array(grid_mapping_table, dtype="int64")

        # Currently 'intc', but will be 'int64' in next major version.
        if int(__version__.split(".")[0]) < 3:
            dtype = "intc"
        else:
            dtype = "int64"

        if self._fit_in_BZ:
            grid_address, _ = relocate_BZ_grid_address(
                grid_address,
                self._mesh,
                self._rec_lat,
                is_shift=self._is_shift,
                is_dense=True,
            )
            self._grid_address = np.array(
                grid_address[: np.prod(self._mesh)], dtype=dtype, order="C"
            )
        else:
            self._grid_address = np.array(grid_address, dtype=dtype, order="C")

        (self._ir_grid_points, self._ir_weights) = extract_ir_grid_points(
            grid_mapping_table
        )

        shift = np.array(self._is_shift) * 0.5
        self._ir_qpoints = np.array(
            (self._grid_address[self._ir_grid_points] + shift) / self._mesh,
            dtype="double",
            order="C",
        )

        self._grid_mapping_table = grid_mapping_table


class GeneralizedRegularGridPoints:
    """Generalized regular grid points.

    Method strategy in suggest mode
    -------------------------------
    1. Create conventional unit cell using spglib.
    2. Sample regular grid points for the conventional unit cell (mesh_numbers)
    3. Transformation matrix from primitive to conventinal unit cell (inv_pmat)
    4. Get supercell multiplicities (mesh_numbers) from the conventional unit
       cell considering the lattice shape.
    5. mmat = (inv_pmat * mesh_numbers).T, which is related to the
       transformation from primitive cell to supercell.
    6. D = P.mmat.Q, where D = diag([n1, n2, n3])
    7. Grid points for primitive cell are
       [np.dot(Q, g) for g in ndindex((n1, n2, n3))].

    Method strategy in non-suggest mode
    -----------------------------------
    1. Find symmetry operations
    2. Determine point group and transformation matrix (tmat) from input cell
    3. Get supercell multiplicities (mesh_numbers) from the transformed cell
       considering the lattice shape.
    4. mmat = (tmat * mesh_numbers).T
    5. D = P.mmat.Q, where D = diag([n1, n2, n3])
    6. Grid points for primitive cell are
       [np.dot(Q, g) for g in ndindex((n1, n2, n3))].

    Attributes
    ----------
    grid_address : ndarray
        Grid addresses in integers.
        shape=(num_grid_points, 3), dtype='int64', order='C'
    qpoints : ndarray
        q-points with respect to basis vectors of input or standardized
        primitive cell.
        shape=(num_grid_points, 3), dtype='double', order='C'
    grid_matrix : ndarray
        Grid generating matrix.
        shape=(3,3), dtype='int64', order='C'
    matrix_to_primitive : ndarray or None
        None when ``suggest`` is False. Otherwise, transformation matrix from
        input cell to the suggested primitive cell.
        shape=(3,3), dtype='double', order='C'
    snf : SNF3x3
        SNF3x3 instance of grid generating matrix.

    """

    def __init__(
        self,
        cell,
        length,
        suggest=True,
        is_time_reversal=True,
        x_fastest=True,
        symprec=1e-5,
    ):
        """Init method.

        Parameters
        ----------
        cell : PhonopyAtoms
            Input cell.
        length : float
            Length having the unit of direct space length.
        suggest : bool, optional, default True
            With True, a standardized primitive cell is suggested and the grids
            are generated for it. With False, input cell is used.
        is_time_reversal: bool, optional, default True
            Time reversal symmetry is considered in symmetry search. By this,
            inversion symmetry is always included.
        x_fastest : bool, optional, default=True
            In grid generation, [[x, y, z], ...], x runs fastest when True,
            otherwise z runs fastest.

        """
        self._cell = cell
        self._length = length
        self._suggest = suggest
        self._is_time_reversal = is_time_reversal
        self._x_fastest = x_fastest
        self._grid_address = None
        self._snf = None
        self._transformation_matrix = None
        self._grid_matrix = None
        self._reciprocal_operations = None
        self._prepare(cell, length, symprec)
        self._generate_grid_points()
        self._generate_q_points()
        self._reciprocal_operations = get_reciprocal_operations(
            self._sym_dataset.rotations,
            self._transformation_matrix,
            self._snf.D,
            self._snf.Q,
            is_time_reversal=self._is_time_reversal,
        )

    @property
    def grid_address(self):
        """Return all grid point addresses."""
        return self._grid_address

    @property
    def qpoints(self):
        """Return all q-points."""
        return self._qpoints

    @property
    def grid_matrix(self):
        """Return grid generating matrix."""
        return self._grid_matrix

    @property
    def transformation_matrix(self):
        """Return transformation matrix."""
        return self._transformation_matrix

    @property
    def snf(self):
        """Return SNF3x3 instance of grid generating matrix."""
        return self._snf

    @property
    def mesh_numbers(self):
        """Return mesh numbers for reciprocal conventional unit cell."""
        return self._mesh_numbers

    @property
    def reciprocal_operations(self):
        """Return reciprocal point group operations."""
        return self._reciprocal_operations

    def _prepare(self, cell, length, symprec):
        """Define grid generating matrix and run the SNF."""
        self._sym_dataset = get_dot_access_dataset(
            get_symmetry_dataset(cell.totuple(), symprec=symprec)
        )
        if self._suggest:
            self._set_grid_matrix_by_std_primitive_cell(cell, length)
        else:
            self._set_grid_matrix_by_input_cell(cell, length)
        self._snf = SNF3x3(self._grid_matrix)
        self._snf.run()

    def _set_grid_matrix_by_std_primitive_cell(self, cell, length):
        """Grid generating matrix based on standardized primitive cell."""
        tmat = self._sym_dataset.transformation_matrix
        centring = self._sym_dataset.international[0]
        pmat = get_primitive_matrix_by_centring(centring)
        conv_lat = np.dot(np.linalg.inv(tmat).T, cell.cell)
        num_cells = np.prod(length2mesh(length, conv_lat))
        self._mesh_numbers = estimate_supercell_matrix(
            self._sym_dataset,
            max_num_atoms=num_cells * len(self._sym_dataset.std_types),
        )
        inv_pmat = np.linalg.inv(pmat)
        inv_pmat_int = np.rint(inv_pmat).astype(int)
        assert (np.abs(inv_pmat - inv_pmat_int) < 1e-5).all()
        # transpose in reciprocal space
        self._grid_matrix = np.array(
            (inv_pmat_int * self._mesh_numbers).T, dtype="int64", order="C"
        )
        # From input lattice to the primitive lattice in real space
        self._transformation_matrix = np.array(
            np.dot(np.linalg.inv(tmat), pmat), dtype="double", order="C"
        )

    def _set_grid_matrix_by_input_cell(self, input_cell, length):
        """Grid generating matrix based on input cell."""
        pointgroup = get_pointgroup(self._sym_dataset.rotations)
        # tmat: From input lattice to point group preserving lattice
        tmat = pointgroup[2]
        lattice = np.dot(input_cell.cell.T, tmat).T
        num_cells = np.prod(length2mesh(length, lattice))
        self._mesh_numbers = estimate_supercell_matrix_from_pointgroup(
            pointgroup[1], lattice, num_cells
        )
        # transpose in reciprocal space
        self._grid_matrix = np.array(
            np.multiply(tmat, self._mesh_numbers).T, dtype="int64", order="C"
        )
        self._transformation_matrix = np.eye(3, dtype="double", order="C")

    def _generate_grid_points(self):
        d = np.diagonal(self._snf.D)
        if self._x_fastest:
            # x runs fastest.
            z, y, x = np.meshgrid(range(d[2]), range(d[1]), range(d[0]), indexing="ij")
        else:
            # z runs fastest.
            x, y, z = np.meshgrid(range(d[0]), range(d[1]), range(d[2]), indexing="ij")
        self._grid_address = np.array(
            np.c_[x.ravel(), y.ravel(), z.ravel()], dtype="int64", order="C"
        )

    def _generate_q_points(self):
        D_inv = np.linalg.inv(self._snf.D)
        qpoints = np.dot(self._grid_address, np.dot(self._snf.Q, D_inv).T)
        qpoints -= np.rint(qpoints)
        self._qpoints = qpoints


def get_reciprocal_operations(
    rotations, transformation_matrix, D, Q, is_time_reversal=True
):
    """Generate reciprocal rotation matrices.

    Collect unique real space rotation matrices and transpose them.
    When is_time_reversal=True, inversion is added if it is not in the
    list of the rotation matrices.

    Parameters
    ----------
    rotations : ndarray
        Rotation matrices in real space. x' = Rx.
        shape=(rotations, 3, 3), dtype='int64'
    transformation_matrxi : array_like
        Transformation matrix of basis vectors in real space. Using this
        rotation matrices are transformed.
    D : array_like
        D of smith normal form 3x3.
        shape=(3, 3)
    Q : array_like
        Q of smith normal form 3x3.
        shape=(3, 3)
    is_time_reversal : bool
        When True, inversion operation is added.

    Returns
    -------
    rotations_for_Q : ndarray
        Rotation matrices in reciprocal space. Grid points are sent by the
        symmetrically equivalent grid points as follows:

        g' = (R_Q g) % diagonal(D)

        shape=(rotations, 3, 3), dtype='int64', order='C'

    """
    unique_rots = []
    tmat_inv = np.linalg.inv(transformation_matrix)
    for r in collect_unique_rotations(rotations):
        _r = similarity_transformation(tmat_inv, r)
        _r_int = np.rint(_r).astype(int)
        assert (np.abs(_r - _r_int) < 1e-5).all()
        unique_rots.append(_r_int)

    ptg_ops, rec_ops = get_pointgroup_operations(
        unique_rots, is_time_reversal=is_time_reversal
    )

    Q_inv = np.linalg.inv(Q)
    rec_ops_Q = []
    for r in rec_ops:
        _r = similarity_transformation(Q_inv, r)
        _r = similarity_transformation(D, _r)
        _r_int = np.rint(_r).astype(int)
        assert (np.abs(_r - _r_int) < 1e-5).all()
        assert abs(determinant(_r_int)) == 1
        rec_ops_Q.append(_r_int)

    return np.array(rec_ops_Q, dtype="int64", order="C")



"""Convert phonon results to animation formats."""

# Copyright (C) 2011 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import numpy as np

from phonopy.interface.vasp import write_vasp
from phonopy.physical_units import get_physical_units
from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.cells import get_angles, get_cell_matrix, get_cell_parameters


def write_animation(
    dynamical_matrix,
    q_point=None,
    anime_type="v_sim",
    band_index=None,
    amplitude=None,
    num_div=None,
    shift=None,
    factor=None,
    filename=None,
):
    """Write atomic modulations in animation format."""
    animation = Animation(q_point, dynamical_matrix, shift=shift)

    if anime_type == "v_sim":
        if filename:
            fname_out = animation.write_v_sim(
                amplitude=amplitude, factor=factor, filename=filename
            )
        else:
            fname_out = animation.write_v_sim(amplitude=amplitude, factor=factor)

    elif anime_type == "arc" or anime_type is None:
        if filename:
            fname_out = animation.write_arc(
                band_index, amplitude, num_div, filename=filename
            )
        else:
            fname_out = animation.write_arc(band_index, amplitude, num_div)
    elif anime_type == "xyz":
        if filename:
            fname_out = animation.write_xyz(
                band_index, amplitude, num_div, factor, filename=filename
            )
        else:
            fname_out = animation.write_xyz(band_index, amplitude, num_div, factor)
    elif anime_type == "jmol":
        if filename:
            fname_out = animation.write_xyz_jmol(
                amplitude=amplitude, factor=factor, filename=filename
            )
        else:
            fname_out = animation.write_xyz_jmol(amplitude=amplitude, factor=factor)
    elif anime_type == "poscar":
        if filename:
            fname_out = animation.write_POSCAR(
                band_index, amplitude, num_div, filename=filename
            )
        else:
            fname_out = animation.write_POSCAR(band_index, amplitude, num_div)
    else:
        raise RuntimeError("Animation format '%s' was not found." % anime_type)

    return fname_out


class Animation:
    """Class to convert phonon results to animation formats."""

    def __init__(self, qpoint, dynamical_matrix, shift=None):
        """Init method."""
        if qpoint is None:
            _qpoint = [0, 0, 0]
        else:
            _qpoint = qpoint
        dynamical_matrix.run(_qpoint)
        dynmat = dynamical_matrix.dynamical_matrix
        self._eigenvalues, self._eigenvectors = np.linalg.eigh(dynmat)
        self._qpoint = _qpoint
        primitive = dynamical_matrix.get_primitive()
        self._positions = primitive.get_scaled_positions()
        self._symbols = primitive.get_chemical_symbols()
        self._masses = primitive.get_masses()
        self._lattice = primitive.get_cell()
        if shift is not None:
            self._positions = (self._positions + shift) % 1

    def _set_cell_oriented(self):
        # Re-oriented lattice xx, yx, yy, zx, zy, zz
        self._angles = get_angles(self._lattice)
        self._cell_params = get_cell_parameters(self._lattice)
        a, b, c = self._cell_params
        alpha, beta, gamma = self._angles
        self._lattice_oriented = get_cell_matrix(a, b, c, alpha, beta, gamma)
        self._positions_oriented = self._get_oriented_displacements(
            np.dot(self._positions, self._lattice)
        )

    # For the orientation, see get_cell_matrix
    def _get_oriented_displacements(self, vec_cartesian):
        return np.dot(
            np.dot(vec_cartesian, np.linalg.inv(self._lattice)), self._lattice_oriented
        )

    def _set_displacements(self, band_index):
        u = []
        for i, e in enumerate(self._eigenvectors[:, band_index]):
            u.append(e / np.sqrt(self._masses[i // 3]))

        self._displacements = np.array(u).reshape(-1, 3)

    def write_v_sim(
        self,
        amplitude=None,
        factor=None,
        filename="anime.ascii",
    ):
        """Write to file in v_sim format."""
        if amplitude is None:
            _amplitude = 1.0
        if factor is None:
            _factor = get_physical_units().DefaultToTHz
        else:
            _factor = factor
        self._set_cell_oriented()
        lat = self._lattice_oriented
        q = self._qpoint
        text = "# Phonopy generated file for v_sim 3.6\n"
        text += "%15.9f%15.9f%15.9f\n" % (lat[0, 0], lat[1, 0], lat[1, 1])
        text += "%15.9f%15.9f%15.9f\n" % (lat[2, 0], lat[2, 1], lat[2, 2])
        for s, p in zip(self._symbols, self._positions_oriented):
            text += "%15.9f%15.9f%15.9f %2s\n" % (p[0], p[1], p[2], s)

        for i, val in enumerate(self._eigenvalues):
            if val > 0:
                omega = np.sqrt(val)
            else:
                omega = -np.sqrt(-val)
            self._set_displacements(i)
            text += "#metaData: qpt=[%f;%f;%f;%f \\\n" % (
                q[0],
                q[1],
                q[2],
                omega * _factor,
            )
            for u in self._get_oriented_displacements(self._displacements) * _amplitude:
                text += "#; %f; %f; %f; %f; %f; %f \\\n" % (
                    u[0].real,
                    u[1].real,
                    u[2].real,
                    u[0].imag,
                    u[1].imag,
                    u[2].imag,
                )
            text += "# ]\n"
        w = open(filename, "w")
        w.write(text)
        w.close()

        return filename

    def write_arc(self, band_index, amplitude=1, num_div=20, filename="anime.arc"):
        """Write to file in BIOSYM archive 3 format."""
        self._set_cell_oriented()
        self._set_displacements(band_index - 1)
        displacements = self._get_oriented_displacements(self._displacements)

        a, b, c = self._cell_params
        alpha, beta, gamma = self._angles

        text = ""
        text += "!BIOSYM archive 3\n"
        text += "PBC=ON\n"

        for i in range(num_div):
            text += "                                                                        0.000000\n"  # noqa E501
            text += "!DATE\n"
            text += "%-4s%10.4f%10.4f%10.4f%10.4f%10.4f%10.4f\n" % (
                "PBC",
                a,
                b,
                c,
                alpha,
                beta,
                gamma,
            )
            positions = (
                self._positions_oriented
                + (displacements * np.exp(2j * np.pi / num_div * i)).imag * amplitude
            )
            for j, p in enumerate(positions):
                text += "%-5s%15.9f%15.9f%15.9f CORE" % (
                    self._symbols[j],
                    p[0],
                    p[1],
                    p[2],
                )
                text += "%5s%3s%3s%9.4f%5s\n" % (
                    j + 1,
                    self._symbols[j],
                    self._symbols[j],
                    0.0,
                    j + 1,
                )

            text += "end\n"
            text += "end\n"

        w = open(filename, "w")
        w.write(text)
        w.close()

        return filename

    def write_xyz_jmol(
        self,
        amplitude=10,
        factor=None,
        filename="anime.xyz_jmol",
    ):
        """Write to file in jmol xyz format."""
        if factor is None:
            _factor = get_physical_units().DefaultToTHz
        else:
            _factor = factor
        self._set_cell_oriented()
        text = ""
        for i, val in enumerate(self._eigenvalues):
            if val > 0:
                freq = np.sqrt(val)
            else:
                freq = -np.sqrt(-val)
            self._set_displacements(i)
            displacements = (
                self._get_oriented_displacements(self._displacements) * amplitude
            )
            text += "%d\n" % len(self._symbols)
            text += "q %s , b %d , f %f " % (str(self._qpoint), i + 1, freq * _factor)
            text += "(generated by Phonopy)\n"
            for s, p, u in zip(self._symbols, self._positions_oriented, displacements):
                text += "%-3s  %22.15f %22.15f %22.15f  " % (s, p[0], p[1], p[2])
                text += "%15.9f %15.9f %15.9f\n" % (u[0].real, u[1].real, u[2].real)
        w = open(filename, "w")
        w.write(text)
        w.close()

        return filename

    def write_xyz(
        self,
        band_index,
        amplitude=1,
        num_div=20,
        factor=None,
        filename="anime.xyz",
    ):
        """Write to file in xyz format."""
        if factor is None:
            _factor = get_physical_units().DefaultToTHz
        else:
            _factor = factor
        self._set_cell_oriented()
        freq = self._eigenvalues[band_index - 1]
        self._set_displacements(band_index - 1)
        displacements = self._get_oriented_displacements(self._displacements)
        text = ""
        for i in range(num_div):
            text += "%d\n" % len(self._symbols)
            text += "q %s , b %d , f %f , " % (
                str(self._qpoint),
                band_index,
                freq * _factor,
            )
            text += "div %d / %d " % (i, num_div)
            text += "(generated by Phonopy)\n"
            positions = (
                self._positions_oriented
                + (displacements * np.exp(2j * np.pi / num_div * i)).imag * amplitude
            )
            for j, p in enumerate(positions):
                text += "%-3s %22.15f %22.15f %22.15f\n" % (
                    self._symbols[j],
                    p[0],
                    p[1],
                    p[2],
                )
        w = open(filename, "w")
        w.write(text)
        w.close()

        return filename

    def write_POSCAR(self, band_index, amplitude=1, num_div=20, filename="APOSCAR"):
        """Write snapshots to files in VASP POSCAR format."""
        self._set_displacements(band_index - 1)
        for i in range(num_div):
            positions = (
                np.dot(self._positions, self._lattice)
                + (self._displacements * np.exp(2j * np.pi / num_div * i)).imag
                * amplitude
            )
            atoms = PhonopyAtoms(
                cell=self._lattice,
                positions=positions,
                masses=self._masses,
                symbols=self._symbols,
            )
            write_vasp((filename + "-%03d") % i, atoms, direct=True)

        return filename



"""Fleur calculator interface."""

# Copyright (C) 2021 Alexander Neukirchen
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import itertools
import math
import sys

import numpy as np

from phonopy.file_IO import collect_forces
from phonopy.interface.vasp import (
    check_forces,
    get_drift_forces,
    sort_positions_by_symbols,
)
from phonopy.structure.atoms import PhonopyAtoms, symbol_map


def parse_set_of_forces(num_atoms, forces_filenames, verbose=True):
    """Parse forces from output files."""
    hook = "1 #"
    is_parsed = True
    force_sets = []

    for i, filename in enumerate(forces_filenames):
        if verbose:
            sys.stdout.write("%d. " % (i + 1))
        f = open(filename)
        fleur_forces = collect_forces(f, num_atoms, hook, [0, 1, 2], word="force")
        if check_forces(fleur_forces, num_atoms, filename, verbose=verbose):
            drift_force = get_drift_forces(
                fleur_forces, filename=filename, verbose=verbose
            )
            force_sets.append(np.array(fleur_forces) - drift_force)
        else:
            is_parsed = False

    if is_parsed:
        return force_sets
    else:
        return []


def read_fleur(filename):
    """Read crystal structure."""
    fleur_in = FleurIn(open(filename).readlines())
    tags = fleur_in.get_variables()
    avec = [tags["avec"][i] for i in range(3)]
    speci = tags["atoms"]["speci"]
    symbols = [
        list(symbol_map.keys())[list(symbol_map.values()).index(math.floor(float(x)))]
        for x in speci
    ]
    numbers = [math.floor(float(x)) for x in speci]

    for i, n in enumerate(numbers):
        if n == 0:
            for j in range(1, 119):
                if j not in numbers:
                    numbers[i] = j
                    break
    pos_all = []
    num_all = []
    for _, pos in zip(numbers, tags["atoms"]["positions"]):
        pos_all += pos
    num_all = [symbol_map[s] for s in symbols]

    return (
        PhonopyAtoms(numbers=num_all, cell=avec, scaled_positions=pos_all),
        speci,
        fleur_in._restlines,
    )


def write_fleur(filename, cell, speci, N, restlines):
    """Write crystal structure to file."""
    f = open(filename, "w")
    f.write(get_fleur_structure(cell, speci, N, restlines))


def write_supercells_with_displacements(
    supercell,
    cells_with_displacements,
    ids,
    speci,
    N,
    restlines,
    pre_filename="supercell",
    width=3,
):
    """Write supercells with displacements to files."""
    write_fleur("%s.in" % pre_filename, supercell, speci, N, restlines)
    for i, cell in zip(ids, cells_with_displacements):
        filename = "{pre_filename}-{0:0{width}}.in".format(
            i, pre_filename=pre_filename, width=width
        )
        write_fleur(filename, cell, speci, N, restlines)


def get_fleur_structure(cell, speci, N, restlines):
    """Return Fleur structure in text."""
    lattice = cell.get_cell()
    (num_atoms, symbols, scaled_positions, sort_list) = sort_positions_by_symbols(
        cell.get_chemical_symbols(), cell.get_scaled_positions()
    )
    specilong = list(
        itertools.chain.from_iterable(itertools.repeat(x, N) for x in speci)
    )
    lines = restlines[0] + "\n"
    lines += ((" %21.16f" * 3 + "\n") * 3) % tuple(lattice.ravel())
    lines += "1.0 \n"
    lines += "1.0 1.0 1.0 \n \n"
    lines += str(sum(num_atoms)) + "\n"
    for i in range(sum(num_atoms)):
        lines += specilong[i].ljust(6)
        currentpos = str(scaled_positions[i]).replace("[", "")
        currentpos = currentpos.replace("]", "").split()

        for j in range(3):
            lines += " " + "{:.10f}".format(float(currentpos[j]))
        if i < sum(num_atoms) - 1:
            lines += "\n"
    for x in range(1, len(restlines)):
        if len(restlines[x]) == 0:
            lines += "\n"
            continue
        lines += restlines[x]
        if x != len(restlines) - 1:
            lines += "\n"
    return lines


class FleurIn:
    """Class to generate Fleur crystal structure."""

    def __init__(self, lines):
        """Init method."""
        self._set_methods = {"a1": self._set_avec, "atoms": self._set_atoms}
        self._tags = {"atoms": None, "a1": None}
        self._lines = lines[:]
        self._restlines = []
        self._collect()

    def get_variables(self):
        """Return variables."""
        return self._tags

    def _collect(self):
        firstline = True
        while True:
            try:
                line_str = self._lines.pop(0).strip()
            except IndexError:
                break

            if firstline is True:
                self._restlines.append(line_str)
                firstline = False
                continue
            if len(line_str) == 0:
                continue
            if line_str[0] == "!":
                continue

            elems = line_str.split()
            if elems[-1] in self._set_methods:
                self._lines.insert(0, line_str)
                self._set_methods[elems[-1]]()

    def _set_atoms(self):
        natoms = int(self._lines.pop(0).split()[0][0])
        speci = []
        positions = []
        positions1 = []
        for _ in range(natoms):
            currentline = self._lines.pop(0).split()
            speci.append(currentline[0])
            currentspeci = [float(x) for x in currentline[1:4]]
            positions1.append(currentspeci)
        positions.append(positions1)

        factors = [1.0, 1.0, 1.0]
        shifts = [0.0, 0.0, 0.0]

        for count, line in enumerate(self._lines):
            if "&factor" in line:
                currentline = self._lines.pop(count).split()
                factors = [
                    float(currentline[1]),
                    float(currentline[2]),
                    float(currentline[3]),
                ]

        for count, line in enumerate(self._lines):
            if "&shift" in line:
                currentline = self._lines.pop(count).split()
                shifts = [
                    float(currentline[1]),
                    float(currentline[2]),
                    float(currentline[3]),
                ]

        for x in positions[0]:
            x[0] = x[0] / factors[0] + shifts[0]
            x[1] = x[1] / factors[1] + shifts[1]
            x[2] = x[2] / factors[2] + shifts[2]

        self._tags["atoms"] = {"speci": speci, "positions": positions}
        for j in range(len(self._lines)):
            self._restlines.append(self._lines[j])

    def _set_avec(self):
        avec = []
        for _ in range(3):
            avec.append([float(x) for x in self._lines.pop(0).split()[:3]])
        lattcon = float(self._lines.pop(0).split()[0])
        scale = [float(x) for x in self._lines.pop(0).split()[:3]]
        for j in range(3):
            for k in range(3):
                if scale[k] < 0:
                    scale[k] = np.sqrt(np.abs(scale[k]))
                if scale[k] == 0.0:
                    scale[k] = 1.0
                avec[j][k] = lattcon * avec[j][k] * scale[k]
        self._tags["avec"] = avec


if __name__ == "__main__":
    from phonopy.structure.symmetry import Symmetry

    cell, speci, restlines = read_fleur(sys.argv[1])
    symmetry = Symmetry(cell)
    print("# %s" % symmetry.get_international_table())
    print(get_fleur_structure(cell, speci, 1, restlines))



"""Mode Grueneisen parameters calculation on sampling mesh."""

# Copyright (C) 2012 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import gzip
import lzma
import sys

import numpy as np

from phonopy.gruneisen.core import GruneisenBase
from phonopy.physical_units import get_physical_units
from phonopy.structure.grid_points import get_qpoints


class GruneisenMesh(GruneisenBase):
    """Class to calculate mode Grueneisen parameters on sampling mesh."""

    def __init__(
        self,
        dynmat,
        dynmat_plus,
        dynmat_minus,
        mesh,
        delta_strain=None,
        shift=None,
        is_time_reversal=True,
        is_gamma_center=False,
        is_mesh_symmetry=True,
        rotations=None,  # Point group operations in real space
        factor=None,
    ):
        """Init method."""
        super().__init__(dynmat, dynmat_plus, dynmat_minus, delta_strain=delta_strain)
        self._mesh = np.array(mesh, dtype="intc")
        if factor is None:
            self._factor = get_physical_units().DefaultToTHz
        else:
            self._factor = factor
        self._cell = dynmat.primitive
        self._qpoints, self._weights = get_qpoints(
            self._mesh,
            np.linalg.inv(self._cell.cell),
            q_mesh_shift=shift,
            is_time_reversal=is_time_reversal,
            is_gamma_center=is_gamma_center,
            rotations=rotations,
            is_mesh_symmetry=is_mesh_symmetry,
        )
        self.set_qpoints(self._qpoints)
        self._gamma = self._gruneisen
        self._frequencies = (
            np.sqrt(abs(self._eigenvalues)) * np.sign(self._eigenvalues) * self._factor
        )

    def get_gruneisen(self):
        """Return mode Grueneisen parameters."""
        return self._gamma

    def get_mesh_numbers(self):
        """Return mesh numbers."""
        return self._mesh

    def get_qpoints(self):
        """Return (irreducible) q-points."""
        return self._qpoints

    def get_weights(self):
        """Return weights of (irreducible) q-points."""
        return self._weights

    def get_eigenvalues(self):
        """Return eigenvalues of dynamical matrices."""
        return self._eigenvalues

    def get_eigenvectors(self):
        """Return phonon eigenvectors."""
        return self._eigenvectors

    def get_frequencies(self):
        """Return phonon frequencies."""
        return self._frequencies

    def write_yaml(self, comment=None, filename=None, compression=None):
        """Write results in yaml file."""
        if filename is not None:
            _filename = filename

        if compression is None:
            if filename is None:
                _filename = "gruneisen.yaml"
            with open(_filename, "w") as w:
                self._write_yaml(w, comment)
        elif compression == "gzip":
            if filename is None:
                _filename = "gruneisen.yaml.gz"
            with gzip.open(_filename, "wb") as w:
                self._write_yaml(w, comment, is_binary=True)
        elif compression == "lzma":
            if filename is None:
                _filename = "gruneisen.yaml.xz"
            with lzma.open(_filename, "w") as w:
                self._write_yaml(w, comment, is_binary=True)

    def _write_yaml(self, w, comment, is_binary=False):
        natom = self._cell.get_number_of_atoms()
        rec_lattice = np.linalg.inv(self._cell.get_cell())  # column vectors
        text = []
        text.append("mesh: [ %5d, %5d, %5d ]" % tuple(self._mesh))
        text.append("nqpoint: %d" % len(self._qpoints))
        text.append("reciprocal_lattice:")
        for vec, axis in zip(rec_lattice.T, ("a*", "b*", "c*")):
            text.append("- [ %12.8f, %12.8f, %12.8f ] # %2s" % (tuple(vec) + (axis,)))
        text.append("natom:   %-7d" % natom)
        text.append(str(self._cell))
        text.append("")
        text.append("phonon:")
        for q, m, gs, freqs in zip(
            self._qpoints, self._weights, self._gamma, self._frequencies
        ):
            text.append("- q-position: [ %10.7f, %10.7f, %10.7f ]" % tuple(q))
            text.append("  multiplicity: %d" % m)
            text.append("  band:")
            for j, (g, freq) in enumerate(zip(gs, freqs)):
                text.append("  - # %d" % (j + 1))
                text.append("    gruneisen: %15.10f" % g)
                text.append("    frequency: %15.10f" % freq)
            text.append("")

        self._write_lines(w, text, is_binary)

    def _write_lines(self, w, lines, is_binary):
        text = "\n".join(lines)
        if is_binary:
            if sys.version_info < (3, 0):
                w.write(bytes(text))
            else:
                w.write(bytes(text, "utf8"))
        else:
            w.write(text)

    def write_hdf5(self, filename="gruneisen.hdf5"):
        """Write results in hdf5 file."""
        import h5py

        w = h5py.File(filename, "w")
        w.create_dataset("mesh", data=self._mesh)
        w.create_dataset("gruneisen", data=self._gamma)
        w.create_dataset("weight", data=self._weights)
        w.create_dataset("frequency", data=self._frequencies)
        w.create_dataset("qpoint", data=self._qpoints)
        w.close()

    def plot(
        self, plt, cutoff_frequency=None, color_scheme=None, marker="o", markersize=None
    ):
        """Return pyplot of calculation results."""
        n = len(self._gamma.T) - 1
        for i, (g, freqs) in enumerate(zip(self._gamma.T, self._frequencies.T)):
            if cutoff_frequency:
                g = np.extract(freqs > cutoff_frequency, g)
                freqs = np.extract(freqs > cutoff_frequency, freqs)

            if color_scheme == "RB":
                color = (1.0 / n * i, 0, 1.0 / n * (n - i))
                if markersize:
                    plt.plot(freqs, g, marker, color=color, markersize=markersize)
                else:
                    plt.plot(freqs, g, marker, color=color)
            elif color_scheme == "RG":
                color = (1.0 / n * i, 1.0 / n * (n - i), 0)
                if markersize:
                    plt.plot(freqs, g, marker, color=color, markersize=markersize)
                else:
                    plt.plot(freqs, g, marker, color=color)
            elif color_scheme == "RGB":
                color = (
                    max(2.0 / n * (i - n / 2.0), 0),
                    min(2.0 / n * i, 2.0 / n * (n - i)),
                    max(2.0 / n * (n / 2.0 - i), 0),
                )
                if markersize:
                    plt.plot(freqs, g, marker, color=color, markersize=markersize)
                else:
                    plt.plot(freqs, g, marker, color=color)
            else:
                if markersize:
                    plt.plot(freqs, g, marker, markersize=markersize)
                else:
                    plt.plot(freqs, g, marker)



"""Mode Grueneisen parameter band structure calculation."""

# Copyright (C) 2012 Atsushi Togo
# All rights reserved.
#
# This file is part of phonopy.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import gzip
import lzma
import sys

import numpy as np
import yaml

from phonopy.gruneisen.core import GruneisenBase
from phonopy.physical_units import get_physical_units


class GruneisenBandStructure(GruneisenBase):
    """Class to calculate mode Grueneisen parameter along band structure paths."""

    def __init__(
        self,
        paths,
        dynmat,
        dynmat_plus,
        dynmat_minus,
        delta_strain=None,
        path_connections=None,
        labels=None,
        factor=None,
    ):
        """Init method."""
        super().__init__(
            dynmat,
            dynmat_plus,
            dynmat_minus,
            delta_strain=delta_strain,
            is_band_connection=True,
        )
        self._cell = dynmat.primitive
        rec_lattice = np.linalg.inv(self._cell.cell)
        distance_shift = 0.0

        if factor is None:
            _factor = get_physical_units().DefaultToTHz
        else:
            _factor = factor

        self._paths = []
        for qpoints_ in paths:
            qpoints = np.array(qpoints_)
            distances = np.zeros(len(qpoints))
            delta_qpoints = qpoints[1:] - qpoints[:-1]
            delta_distances = np.sqrt(
                (np.dot(delta_qpoints, rec_lattice) ** 2).sum(axis=1)
            )
            for i, dd in enumerate(delta_distances):
                distances[i + 1] = distances[i] + dd

            self.set_qpoints(qpoints)
            eigenvalues = self._eigenvalues
            frequencies = np.sqrt(abs(eigenvalues)) * np.sign(eigenvalues) * _factor
            distances_with_shift = distances + distance_shift

            self._paths.append(
                [
                    qpoints,
                    distances,
                    self._gruneisen,
                    eigenvalues,
                    self._eigenvectors,
                    frequencies,
                    distances_with_shift,
                ]
            )

            distance_shift = distances_with_shift[-1]
        self._labels = None
        self._path_connections = None
        if path_connections is None:
            self._path_connections = [
                True,
            ] * len(self._paths)
            self._path_connections[-1] = False
        else:
            self._path_connections = path_connections
        if (
            labels is not None
            and len(labels) == (2 - np.array(self._path_connections)).sum()
        ):
            self._labels = labels

    def get_qpoints(self):
        """Return q-points."""
        return [path[0] for path in self._paths]

    def get_distances(self):
        """Return distances."""
        return [path[6] for path in self._paths]

    def get_gruneisen(self):
        """Return mode Gruneisen parameters."""
        return [path[2] for path in self._paths]

    def get_eigenvalues(self):
        """Return eigenvalues."""
        return [path[3] for path in self._paths]

    def get_eigenvectors(self):
        """Return eigenvectors."""
        return [path[4] for path in self._paths]

    def get_frequencies(self):
        """Return frequencies."""
        return [path[5] for path in self._paths]

    def write_yaml(self, comment=None, filename=None, compression=None):
        """Write results to file in yaml."""
        if filename is not None:
            _filename = filename

        if compression is None:
            if filename is None:
                _filename = "gruneisen.yaml"
            with open(_filename, "w") as w:
                self._write_yaml(w, comment)
        elif compression == "gzip":
            if filename is None:
                _filename = "gruneisen.yaml.gz"
            with gzip.open(_filename, "wb") as w:
                self._write_yaml(w, comment, is_binary=True)
        elif compression == "lzma":
            if filename is None:
                _filename = "gruneisen.yaml.xz"
            with lzma.open(_filename, "w") as w:
                self._write_yaml(w, comment, is_binary=True)

    def _write_yaml(self, w, comment, is_binary=False):
        natom = self._cell.get_number_of_atoms()
        rec_lattice = np.linalg.inv(self._cell.cell)  # column vecs
        nq_paths = []
        for qpoints in self._paths:
            nq_paths.append(len(qpoints))
        text = []
        if comment is not None:
            text.append(yaml.dump(comment, default_flow_style=False).rstrip())
        text.append("nqpoint: %-7d" % np.sum(nq_paths))
        text.append("npath: %-7d" % len(self._paths))
        text.append("segment_nqpoint:")
        text += ["- %d" % nq for nq in nq_paths]
        if self._labels:
            text.append("labels:")
            if self._is_legacy_plot:
                for i in range(len(self._paths)):
                    text.append(
                        "- [ '%s', '%s' ]" % (self._labels[i], self._labels[i + 1])
                    )
            else:
                i = 0
                for c in self._path_connections:
                    text.append(
                        "- [ '%s', '%s' ]" % (self._labels[i], self._labels[i + 1])
                    )
                    if c:
                        i += 1
                    else:
                        i += 2
        text.append("reciprocal_lattice:")
        for vec, axis in zip(rec_lattice.T, ("a*", "b*", "c*")):
            text.append("- [ %12.8f, %12.8f, %12.8f ] # %2s" % (tuple(vec) + (axis,)))
        text.append("natom: %-7d" % (natom))
        text.append(str(self._cell))
        text.append("")
        text.append("path:")
        text.append("")

        for band_structure in self._paths:
            (
                qpoints,
                distances,
                gamma,
                eigenvalues,
                _,
                frequencies,
                distances_with_shift,
            ) = band_structure

            text.append("- nqpoint: %d" % len(qpoints))
            text.append("  phonon:")
            for q, d, gs, freqs in zip(qpoints, distances, gamma, frequencies):
                text.append("  - q-position: [ %10.7f, %10.7f, %10.7f ]" % tuple(q))
                text.append("    distance: %10.7f" % d)
                text.append("    band:")
                for i, (g, freq) in enumerate(zip(gs, freqs)):
                    text.append("    - # %d" % (i + 1))
                    text.append("      gruneisen: %15.10f" % g)
                    text.append("      frequency: %15.10f" % freq)
                text.append("")

        self._write_lines(w, text, is_binary)

    def _write_lines(self, w, lines, is_binary):
        text = "\n".join(lines)
        if is_binary:
            if sys.version_info < (3, 0):
                w.write(bytes(text))
            else:
                w.write(bytes(text, "utf8"))
        else:
            w.write(text)

    def plot(self, axarr, epsilon=None, color_scheme=None):
        """Return pyplot of band structure calculation results."""
        for band_structure in self._paths:
            self._plot(axarr, band_structure, epsilon, color_scheme)

    def _plot(self, axarr, band_structure, epsilon, color_scheme):
        (
            qpoints,
            distances,
            gamma,
            eigenvalues,
            _,
            frequencies,
            distances_with_shift,
        ) = band_structure

        n = len(gamma.T) - 1
        ax1, ax2 = axarr

        for i, (curve, freqs) in enumerate(zip(gamma.T.copy(), frequencies.T)):
            if epsilon is not None:
                if np.linalg.norm(qpoints[0]) < epsilon:
                    cutoff_index = 0
                    for j, q in enumerate(qpoints):
                        if not np.linalg.norm(q) < epsilon:
                            cutoff_index = j
                            break
                    for j in range(cutoff_index):
                        if abs(freqs[j]) < abs(max(freqs)) / 10:
                            curve[j] = curve[cutoff_index]

                if np.linalg.norm(qpoints[-1]) < epsilon:
                    cutoff_index = len(qpoints) - 1
                    for j in reversed(range(len(qpoints))):
                        q = qpoints[j]
                        if not np.linalg.norm(q) < epsilon:
                            cutoff_index = j
                            break
                    for j in reversed(range(len(qpoints))):
                        if j == cutoff_index:
                            break
                        if abs(freqs[j]) < abs(max(freqs)) / 10:
                            curve[j] = curve[cutoff_index]

            self._plot_a_band(ax1, curve, distances_with_shift, i, n, color_scheme)
        ax1.set_xlim(0, distances_with_shift[-1])

        for i, freqs in enumerate(frequencies.T):
            self._plot_a_band(ax2, freqs, distances_with_shift, i, n, color_scheme)
        ax2.set_xlim(0, distances_with_shift[-1])

    def _plot_a_band(self, ax, curve, distances_with_shift, i, n, color_scheme):
        color = None
        if color_scheme == "RB":
            color = (1.0 / n * i, 0, 1.0 / n * (n - i))
        elif color_scheme == "RG":
            color = (1.0 / n * i, 1.0 / n * (n - i), 0)
        elif color_scheme == "RGB":
            color = (
                max(2.0 / n * (i - n / 2.0), 0),
                min(2.0 / n * i, 2.0 / n * (n - i)),
                max(2.0 / n * (n / 2.0 - i), 0),
            )
        if color:
            ax.plot(distances_with_shift, curve, color=color)
        else:
            ax.plot(distances_with_shift, curve)
